r"""
Chiodo class from JPPZ - Corollary 4
"""
from __future__ import absolute_import, print_function

import itertools

from six.moves import range

from admcycles.admcycles import Tautv_to_tautclass, Tautvb_to_tautclass, Tautvecttobasis, tautgens, psiclass, tautclass, list_strata
from admcycles.stable_graph import StableGraph
from admcycles.graph_sum import graph_sum, divbyaut
from admcycles.double_ramification_cycle
import admcycles.DR as DR

from sage.combinat.subset import Subsets
from sage.arith.all import factorial
from sage.functions.other import floor, ceil
from sage.misc.misc_c import prod
from sage.rings.all import PolynomialRing, QQ, ZZ
from sage.modules.free_module_element import vector
from sage.rings.polynomial.polynomial_ring import polygen
from sage.rings.polynomial.multi_polynomial_element import MPolynomial

def chiodo_class(g, Avector, d=None, k=None, rpoly=False, tautout=True, basis=False):
    if d is None:
        d = g
    n = len(Avector)

    if any(isinstance(ai, MPolynomial) for ai in Avector):
        # compute DRpoly and substitute the ai in there
        pol, gens = DRpoly(g, d, n, tautout=tautout, basis=basis, gensout=True)
        subsdict = {gens[i]: Avector[i] for i in range(n)}

        if tautout:
            pol.coeff_subs(subsdict)
            pol.simplify()
            return pol
        return pol.apply_map(lambda x: x.subs(subsdict))

    if k is None:
        k = floor(sum(Avector)/(2*g-2+n))
    if sum(Avector) != k*(2*g-2+n):
        raise ValueError('2g-2+n must divide the sum of entries of Avector.')

    gens = tautgens(g, n, d, decst=True)
    v = vector([DR_coeff_new(decs, g, n, d, Avector, k, rpoly) for decs in gens]) / ZZ(2)**d
    #v1=vector([QQ(a) for a in DR.convert_vector_to_monomial_basis(v,g,d,tuple(range(1, n+1)),DR.MODULI_ST)])

    if basis:
        v1 = Tautvecttobasis(v, g, n, d)
        if tautout:
            return Tautvb_to_tautclass(v1, g, n, d)
        return v1
    else:
        if tautout:
            return Tautv_to_tautclass(v, g, n, d)
        return v


def DR_coeff_setup(G, g, n, d, Avector, k):
    gamma = G.gamma
    kappa, psi = G.poly.monom[0]
    exp_list = []
    scalar_factor = QQ((1, G.automorphism_number()))
    given_weights = [-k * (2 * gv - 2 + len(gamma._legs[i]))
                     for i, gv in enumerate(gamma._genera)]

    # contributions to scalar_factor from kappa-classes
    for v in range(gamma.num_verts()):
        if len(kappa[v]) == 1: # some kappa_1^e at this vertex
            scalar_factor *= (-k**2)**(kappa[v][0]) / factorial(kappa[v][0])

    # contributions to scalar_factor and given_weights from markings
    for i in range(1, n + 1):
        v = gamma.vertex(i)
        psipow = psi.get(i, 0) # if i in dictionary, have psi_i^(psi[i]), otherwise have psi_i^0
        given_weights[v] += Avector[i - 1]
        scalar_factor *= (Avector[i - 1])**(2 * psipow) / factorial(psipow)

    # contributions to scalar_factor and explist from edges
    for (lu, lv) in gamma._edges:
        psipowu = psi.get(lu, 0)
        psipowv = psi.get(lv, 0)
        exp_list.append(psipowu + psipowv + 1)
        scalar_factor *= ((-1)**(psipowu+psipowv)) / factorial(psipowv) / factorial(psipowu) / (psipowu+psipowv+1)

    return exp_list, given_weights, scalar_factor


def DR_coeff_new(G, g, n, d, Avector, k, rpoly):
    gamma = G.gamma # underlying stable graph of decstratum G
    kappa, _ = G.poly.monom[0]
    # kappa is a list of lenght = # vertices, of entries like [3,0,2] meaning that at this vertex there is a kappa_1^3 * kappa_3^2
    # _ = psi is a dictionary and psi[3]=4 means there is a decoration psi^4 at marking/half-edge 3

    if any(len(kalist) > 1 for kalist in kappa):
        return 0  # vertices can only carry powers of kappa_1, no higher kappas allowed

    m0 = ceil(sum([abs(i) for i in Avector]) / ZZ(2)) + g*abs(k) # value from which on the Pixton-sum below will be polynomial in m
    h0 = gamma.num_edges() - gamma.num_verts() + 1 # first Betti number of the graph Gamma
    exp_list, given_weights, scalar_factor = DR_coeff_setup(G, g, n, d, Avector, k)

    deg = 2 * sum(exp_list)  # degree of polynomial in m
    # R = PolynomialRing(QQ, len(gamma.edges) + 1, 'd')
    # P=prod([(di*(d0-di))**exp_list[i] for i,di in enumerate(R.gens()[1:])])/(d0**h0)
    u = gamma.flow_solve(given_weights)
    Vbasis = gamma.cycle_basis()

    #mpoly = mpoly_affine_sum(P, u, Vbasis,m0+1,deg)
    mpoly = mpoly_special(exp_list, h0, u, Vbasis, m0+1, deg)

    # mpoly = interpolate(mrange, mvalues)
    if rpoly:
        return mpoly * scalar_factor
    return mpoly.subs(r=0) * scalar_factor


def mpoly_special(exp_list, h0, u, Vbasis, start, degr):
    r"""
    Return the sum of the rational function in DR_coeff_new over the
    affine space u + V modulo r in ZZ^n as a univariate polynomial in r.
    V is the lattice with basis Vbasis.
    Use that this sum is polynomial in r starting from r=start and the polynomial has
    degree degr (for now).
    """
    mrange = list(range(start, start + degr + 2))
    mvalues = []
    rank = len(Vbasis)
    ulen = len(u)

    for m in mrange:
        total = 0
        for coeff in itertools.product(*[list(range(m)) for i in range(rank)]):
            v = u + sum([coeff[i]*Vbasis[i] for i in range(rank)])
            v = [vi % m for vi in v]
            total += prod([(v[i]*(m-v[i]))**exp_list[i] for i in range(ulen)])
        mvalues.append(total/QQ(m**h0))

    return interpolate(mrange, mvalues)



"""
def chiodo_class(g, Avector, d=None, k=None, r=None, rpoly=False, tautout=True, basis=False):
    r"""Returns the formula epsilon_* c(- R* pi_* L) from [JPPZ, Corollary 4]."""
    if d is None:
        d = g
    n = len(Avector)
    if k is None:
        k = floor(sum(Avector)/(2*g-2+n))
    if sum(Avector) != k*(2*g-2+n):
        raise ValueError('2g-2+n must divide the sum of entries of Avector.')
    if r == None and not rpoly:
        raise ValueError('Either r must be specified or rpoly must be set to True.')        
    
    if not r is None:
        GWlist = GammaWlist(g, Avector, d, k, r)
        GF = GFactor(r)
        
        return graph_sum(g,n,decgraphs=GWlist,globalfact=GF, vertterm=None,legterm=None,edgeterm=None,maxdeg=None,deg=None,termsout=False)
                

def GammaWlist(g, Avector, d, k, r):
    n = len(Avector)
    for e in range(d+1): # number of edges"""
        for gamma in list_strata(g, n, e):
            



from __future__ import print_function

import itertools

from sympy.utilities.iterables import partitions, permutations, multiset_partitions

# pylint does not know sage
from sage.graphs.graph import Graph # pylint: disable=import-error
from sage.graphs.generators.basic import CompleteBipartiteGraph # pylint: disable=import-error
# sage documentation says compositions are better than OrderedPartitions, no idea why....
from sage.combinat.composition import Compositions as sage_part # pylint: disable=import-error
from sage.combinat.partition import OrderedPartitions # pylint: disable=import-error
from sage.misc.cachefunc import cached_function # pylint: disable=import-error

from admcycles.stable_graph import StableGraph as stgraph
import admcycles.diffstrata.levelgraph
import admcycles.diffstrata.levelstratum
from admcycles.diffstrata.sig import Signature

############################################################
############################################################
#### Old BIC generation:
#### This was the first attempt at generating BICs 
#### (generating all stable graphs with admcycles and then
#### putting all possible bipartite structures etc on them).
#### It's nice for historic reasons but painfully slow!
#### Definitely use bic_alt below instead!!!
############################################################
############################################################

def SageGraph(gr):
    # converts stgraph gr in Sage graph
    legdic={i:vnum for vnum in range(len(gr.legs(copy=False))) for i in gr.legs(vnum,copy=False)}
    return Graph([(legdic[a], legdic[b]) for (a,b) in gr.edges(copy=False)],loops=True, multiedges=True)

def bics(g,orders):
    """
    Generate BICs for stratum with signature orders of genus g.

    DO NOT USE!!! USE bic_alt instead!!!
    
    Args:
        g (int): genus
        orders (tuple): signature tuple
    
    Returns:
        list: list of BICs
    """
    print('WARNING! This is not the normal BIC generation algorithm!')
    print("Only use if you know what you're doing!")
    n=len(orders)
    bound=g+n
    result=[] # list of levelgraphs
    orderdict={i+1:orders[i] for i in range(n)}
  
    for e in range(1,bound+1):
        # look at which stgraphs in Mbar_g,n with e edges are bipartite
        for gr in new_list_strata(g,n,e):
            check=SageGraph(gr).is_bipartite(True)
            if check[0]:
                # check[1] contains dictionary {vertices: 0 or 1}
                vert1=[v for v in check[1] if check[1][v]==0]
                vert2=[v for v in check[1] if check[1][v]==1]
                result+=bics_on_bipgr(gr,vert1,vert2,orderdict)
                result+=bics_on_bipgr(gr,vert2,vert1,orderdict)
    return result
    
def bics_on_bipgr(gr,vertup,vertdown,orderdict):
    # takes stable graph and partition of range(len(gr.genera)) into vertup, vertdown
    # creates list of possible pole orders at edges
    result=[]

    # removed from admcycles so reinserted here for legacy purposes:
    def dicunion(*dicts):
        return dict(itertools.chain.from_iterable(dct.items() for dct in dicts))
    
    numvert=len(gr.genera(copy=False))
    halfedges=gr.halfedges()
    halfedgeinvers={e0:e1 for (e0,e1) in gr.edges(copy=False)}
    halfedgeinvers.update({e1:e0 for (e0,e1) in gr.edges(copy=False)})
    
    levels=[0 for i in range(numvert)]
    for i in vertdown:
        levels[i]=-1
    
    helist=[[l for l in ll if l in halfedges] for ll in gr.legs(copy=False)] #list of half-edges of each vertex
    degs=[] # list (with len numvert) of orders that need to be provided by pole orders of half-edges (not legs)
    for v in range(numvert):
        degs.append(2*gr.genera(v)-2-sum([orderdict[l] for l in gr.list_markings(v)]))
        # if degs<0:
        #     return []
    
    weightparts=[]
    for v in vertup:
        vweightpart=[]
        for part in OrderedPartitions(degs[v]+len(helist[v]),len(helist[v])):
            # dictionary of weights from edges connected to vertex v
            vdic={helist[v][i]:part[i]-1 for i in range(len(helist[v]))}
            vdic.update({halfedgeinvers[helist[v][i]]:-part[i]-1 for i in range(len(helist[v]))})
            #print vdic
            vweightpart.append(vdic)
        weightparts+=[vweightpart]
        #print weightparts
    
    for parts in itertools.product(*weightparts):
        #print parts
        poleorders=dicunion(orderdict,*parts)
        CandGraph=admcycles.diffstrata.levelgraph.LevelGraph(gr.genera(copy=False),gr.legs(copy=False),gr.edges(copy=False),poleorders,levels,quiet=True)
        if CandGraph.is_legal(True) and CandGraph.checkadmissible(True):
            result.append(CandGraph)
    return result
    
def new_list_strata(g,n,r):
    return [lis[0] for lis in admcycles.admcycles.degeneration_graph(int(g),n,r)[0][r] ]

############################################################
############################################################

############################################################
############################################################
#### New BIC generation. This is currently used.
############################################################
############################################################

@cached_function
def bic_alt_noiso(sig):
    """
    Construct all non-horizontal divisors in the stratum sig.

    More precisely, each BIC is LevelGraph with two levels numbered 0, -1
    and marked points 1,...,n where the i-th point corresponds to the element
    i-1 of the signature.

    Note that this is the method called by GeneralisedStratum.gen_bic.
    
    Args:
        sig (Signature): Signature of the stratum.
    
    Returns:
        list: list of 2-level non-horizontal LevelGraphs.
 
    EXAMPLES ::

        sage: from admcycles.diffstrata import *
        sage: assert comp_list(bic_alt(Signature((1,1))),\
        [LevelGraph([1, 0],[[3, 4], [1, 2, 5, 6]],[(3, 5), (4, 6)],{1: 1, 2: 1, 3: 0, 4: 0, 5: -2, 6: -2},[0, -1],True),\
        LevelGraph([1, 1, 0],[[4], [3], [1, 2, 5, 6]],[(3, 5), (4, 6)],{1: 1, 2: 1, 3: 0, 4: 0, 5: -2, 6: -2},[0, 0, -1],True),\
        LevelGraph([1, 1],[[3], [1, 2, 4]],[(3, 4)],{1: 1, 2: 1, 3: 0, 4: -2},[0, -1],True),\
        LevelGraph([2, 0],[[3], [1, 2, 4]],[(3, 4)],{1: 1, 2: 1, 3: 2, 4: -4},[0, -1],True)])

        sage: assert comp_list(bic_alt(Signature((2,))),\
        [LevelGraph([1, 1],[[2], [1, 3]],[(2, 3)],{1: 2, 2: 0, 3: -2},[0, -1],True),\
        LevelGraph([1, 0],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 2, 2: 0, 3: 0, 4: -2, 5: -2},[0, -1],True)])

        sage: assert comp_list(bic_alt(Signature((4,))),\
        [LevelGraph([1, 1, 0],[[2, 4], [3], [1, 5, 6, 7]],[(2, 5), (3, 6), (4, 7)],{1: 4, 2: 0, 3: 0, 4: 0, 5: -2, 6: -2, 7: -2},[0, 0, -1],True),\
        LevelGraph([1, 1, 1],[[3], [2], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 0, 3: 0, 4: -2, 5: -2},[0, 0, -1],True),\
        LevelGraph([2, 0],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 2, 3: 0, 4: -4, 5: -2},[0, -1],True),\
        LevelGraph([2, 0],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 1, 3: 1, 4: -3, 5: -3},[0, -1],True),\
        LevelGraph([1, 0],[[2, 3, 4], [1, 5, 6, 7]],[(2, 5), (3, 6), (4, 7)],{1: 4, 2: 0, 3: 0, 4: 0, 5: -2, 6: -2, 7: -2},[0, -1],True),\
        LevelGraph([1, 2],[[2], [1, 3]],[(2, 3)],{1: 4, 2: 0, 3: -2},[0, -1],True),\
        LevelGraph([2, 1],[[2], [1, 3]],[(2, 3)],{1: 4, 2: 2, 3: -4},[0, -1],True),\
        LevelGraph([1, 1],[[2, 3], [1, 4, 5]],[(2, 4), (3, 5)],{1: 4, 2: 0, 3: 0, 4: -2, 5: -2},[0, -1],True)])

        sage: len(bic_alt(Signature((1,1,1,1)))) # long time (2 seconds)
        102

        sage: len(bic_alt(Signature((2,2,0,-2))))
        61

        sage: len(bic_alt((2,2,0,-2)))
        61
    """
    if isinstance(sig, tuple):
        sig = Signature(sig)
    zeros = [i+1 for i,a in enumerate(sig.sig) if a > 0] # for keeping track of the numbered half-edges, we remember the index
    z = sig.z
    poles = [i+1 for i,a in enumerate(sig.sig) if a < 0]
    p = sig.p
    marked_points = [i+1 for i,a in enumerate(sig.sig) if a == 0]
    mp = len(marked_points)
    g = sig.g
    orders = {i+1 : ord for i,ord in enumerate(sig.sig)}

    n = sig.n

    # As every g=0 top component needs at least one pole, the bottom max depends on this:
    if p > 0:
        g_bot_max = g
        g_top_min = 0
        pole_ind = [0,1]
    else:
        g_bot_max = g - 1
        g_top_min = 1
        pole_ind = [0] # poles don't need to be distributed...
    
    found = []

    for bot_comp_len in range(1,z+mp+1):
        # every component on bottom level needs at least one zero or two marked points 
        # (in case it's genus 0 and has only one edge with a double pole going up...)
        # Therefore, we split the zeros into top and bottom and distribute these:
        for parts in itertools.chain(multiset_partitions(zeros,2), iter([[zeros,[]]])):
            for i in [0,1]:
                if mp > 0 or len(parts[i]) >= bot_comp_len:
                    bottom_zeros = parts[i]
                    top_zeros = parts[1-i]
                else:
                    continue
                # if there are no marked points, every component needs a zero
                # otherwise we are more flexible (this could be improved, but is good enoug for now)
                if mp == 0:
                    bot_zero_gen = _distribute_fully(bottom_zeros,bot_comp_len)
                else:
                    bot_zero_gen = _distribute_points(bottom_zeros,bot_comp_len)
                for distr_bot_zeros in bot_zero_gen:
                    # partition genus into top, bottom and graph:
                    for total_g_bot in range(g_bot_max+1):
                        for total_g_top in range(g_top_min, g - total_g_bot + 1):
                            total_g_graph = g - total_g_bot - total_g_top
                            # partition bottom genus into components
                            for distr_bot_g in _distribute_part_ordered(total_g_bot,bot_comp_len):
                                # first test:
                                # orders on each component must add up to 2g-2, from now on only poles get
                                # added and we need at least one edge going up (adding at least a pole of order 2)
                                # So if sum(zeros) < 2g, we're screwed
                                if not all(sum(orders[bz] for bz in distr_bot_zeros[c]) 
                                            >= 2*distr_bot_g[c] for c in range(bot_comp_len)):
                                    continue
                                # start distributing poles, first between top and bottom:
                                for pole_parts in itertools.chain(multiset_partitions(poles,2), iter([[poles,[]]])):
                                    for ip in pole_ind: # no poles => nothing to distribute
                                        bottom_poles = pole_parts[ip]
                                        top_poles = pole_parts[1-ip]
                                        # poles on top necessary for g=0 components
                                        if total_g_top == 0 and len(top_poles) == 0:
                                            continue
                                        # poles are optional (i.e. not every vertex needs a pole)
                                        for distr_bot_poles in _distribute_points(bottom_poles,bot_comp_len):
                                            # we save the "space" left to distribute among poles from half-edges:
                                            spaces_bot = [(-sum(orders[tz] for tz in distr_bot_zeros[c]) 
                                                        - sum(orders[tp] for tp in distr_bot_poles[c]) 
                                                                + 2*distr_bot_g[c] - 2) for c in range(bot_comp_len)]
                                            # retest: each space must be at least -2
                                            if not all(s <= -2 for s in spaces_bot):
                                                continue
                                            # iterate over top components
                                            for top_comp_len in range(1,total_g_top+len(top_poles)+1):
                                                # now we know all vertices and the genus distribution, we know the number of edges:
                                                num_of_edges = top_comp_len + bot_comp_len + total_g_graph - 1
                                                # "global" condition on bottom: 
                                                # for each edge there will be at least another pole of order 2
                                                if (sum(sum(orders[bz] for bz in distr_bot_zeros[c]) 
                                                        + sum(orders[bp] for bp in distr_bot_poles[c]) 
                                                            for c in range(bot_comp_len))
                                                        - 2*num_of_edges < 2*total_g_bot - 2*bot_comp_len):
                                                    continue
                                                # distribute genus and poles
                                                for distr_top_g in _distribute_part_ordered(total_g_top,top_comp_len):
                                                    for distr_top_poles in _distribute_points(top_poles,top_comp_len):
                                                        # test: if orders add up to more than 2g-2 this won't work
                                                        # (effectively this should only concern g=0 components here...)
                                                        if not all(sum(orders[tp] for tp in distr_top_poles[c]) 
                                                                            <= 2*distr_top_g[c] - 2
                                                                                    for c in range(top_comp_len)):
                                                            continue
                                                        # distribute remaining zeros optionally
                                                        for distr_top_zeros in _distribute_points(top_zeros,top_comp_len):
                                                            # again, we record the spaces:
                                                            spaces_top = [-sum(orders[tp] for tp in distr_top_poles[c]) 
                                                                          - sum(orders[tz] for tz in distr_top_zeros[c])
                                                                                 + 2*distr_top_g[c] - 2
                                                                                        for c in range(top_comp_len)] # yapf: disable
                                                            # retest:
                                                            if not all(s >= 0 for s in spaces_top):
                                                                continue
                                                            # We distribute the edges together with their prongs/poleorders
                                                            # before we check for connectednes
                                                            # We start on bottom level, because there each half-edge
                                                            # comes with a mandatory pole:
                                                            for half_edges_bot, orders_he in _place_legs_on_bot(spaces_bot, num_of_edges, n+num_of_edges):
                                                                # note for the numbering of half-edges:
                                                                # * the HE on bot are numbered n+num_of_edges+1,...,n+2*num_of_edges
                                                                # * the HE on top are numbered the same for now, i.e. edges are (l,l) ,
                                                                #   (and will be renumbered n+1,...,n+num_of_edges in a moment)
                                                                for half_edges_top in _place_legs_on_top(spaces_top,orders_he):
                                                                    # check if graph is connected!
                                                                    # We create the associated sage graph
                                                                    # this doesn't have multiple edges, but we only
                                                                    # care about connectedness
                                                                    # For this, we build a dictionary {leg: comp bottom}
                                                                    edge_dict = {}
                                                                    for j, c in enumerate(half_edges_bot):
                                                                        for l in c:
                                                                            # note that the vertices of G will be numbered
                                                                            # 0,...,top_comp_len+bot_comp_len-1
                                                                            # so the bot components are numbered by their index
                                                                            # offset by top_comp_len
                                                                            edge_dict[l] = j + top_comp_len
                                                                            # add in the orders for top edges (note the offset)
                                                                            orders_he[l-num_of_edges] = -2 - orders_he[l]
                                                                    G = Graph([(j,edge_dict[l]) for j,c in enumerate(half_edges_top)
                                                                                                        for l in c], multiedges=False)
                                                                    if not G.is_connected():
                                                                        continue
                                                                    genera = distr_top_g+distr_bot_g
                                                                    # Distribute order 0 marked points
                                                                    for distr_mp in _distribute_points(marked_points,len(genera)):
                                                                        # combine nested lists of legs:
                                                                        legs = ([distr_top_zeros[c] + distr_top_poles[c]
                                                                                    + distr_mp[c]
                                                                                    # renumber top HE
                                                                                    + [l-num_of_edges for l in half_edges_top[c]]
                                                                                    for c in range(top_comp_len)] # top component
                                                                                + [distr_bot_zeros[c] + distr_bot_poles[c]
                                                                                    + distr_mp[c+top_comp_len] # offset
                                                                                    + half_edges_bot[c]
                                                                                    for c in range(bot_comp_len)] # bot component
                                                                            ) # yapf: disable
                                                                        # check stability:
                                                                        if any(len(ls) < 3 for c,ls in enumerate(legs) 
                                                                                                if genera[c] == 0):
                                                                            continue
                                                                        lg_data = (genera, legs, 
                                                                            [(l,l+num_of_edges) for l in range(n+1,n+num_of_edges+1)],
                                                                            _merge_dicts(orders,orders_he), # orders as dict
                                                                            [0]*top_comp_len + [-1]*bot_comp_len, # levels
                                                                            # True # quiet
                                                                        )
                                                                        LG = admcycles.diffstrata.levelgraph.LevelGraph(*lg_data)
                                                                        if LG.checkadmissible(quiet=True):
                                                                            if LG.is_legal(quiet=True):
                                                                                found.append(LG)
                                                                        else:
                                                                            # this should not happen!
                                                                            print("Not admissible(!): ", LG)
    found = list(set(found)) # remove duplicates
    return found

def bic_alt(sig):
    """
    The BICs of the stratum sig up to isomorphism of LevelGraphs.

    This should not be used directly, use a GeneralisedStratum or Stratum
    instead to obtain EmbeddedLevelGraphs and the correct isomorphism classes.
    
    Args:
        sig (tuple): signature tuple
    
    Returns:
        list: list of LevelGraphs.
    """
    return isom_rep(bic_alt_noiso(sig)) # remove duplicates up to iso

def _merge_dicts(x,y):
    z = x.copy()
    z.update(y)
    return z

def _place_legs_on_bot(space,num_of_points,start):
    ## iterator distributing n legs with pole orders on (bottom) components according to space (each pole order is at least -2)
    ## here space is a list of 2g-2-stuff we already placed on each bottom component
    ## return a pair (pointlist,orderdict) where
    ## * pointlist is a nested list of points (numbered start+1,...,n)
    ## * orderdict is a dictionary leg number : order
    legal_splits = []
    distr = [] # list to hold the current distribution
    def split(n):
        ## distribute n points onto space and add to distr
        # note that space and distr are accessed from back to front...
        if n==0:
            # done
            # correct signs and reverse order
            legal_splits.append([[-a for a in c] for c in reversed(distr)])
            return
        else:
            if not space: # check if empty
                return
            current = space.pop()
            remaining_comp = len(space)
            # there is a sign issue here: partitions are of positive integers, our pole orders are negative
            # each leg has pole order at least 2, so we only list these partitions
            # moreover, we may place at most n-remaining_comp points on this component
            possibilities = sage_part(-current,min_part=2,max_length=n-remaining_comp).list()
            if possibilities: # check if non-empty
                for possibility in possibilities:
                    distr.append(possibility)
                    # recursion
                    split(n-len(possibility))
                    # undo damage
                    distr.pop()
            space.append(current)
    # generate splits:
    split(num_of_points)
    # we now convert the generated lists (of pole orders) into the format: nested list of numbered legs
    # Note: 
    # * the legs are numbered starting from start+1 to start+num_of_points+1
    # * the orders are saved in the list of dictionaries edge_orders_bot
    for dist in legal_splits:
        order_dic = {} 
        p = start + 1
        for c in range(len(dist)):
            for a in range(len(dist[c])):
                order_dic[p] = dist[c][a]
                dist[c][a] = p
                p += 1
        yield (dist, order_dic)

def _place_legs_on_top(space,orders_bot):
    ## iterator distributing n legs with zero orders (determined by their friends on the bottom component)
    ## onto the top components according to space
    ## here space is a list of 2g_comp-2-stuff we already placed on each top component
    ## return a pointlist (numbered according to keys of orders_bot for now...)
    legal_splits = []
    distr = [[] for _ in space] # list to hold current distribution
    # we sort the points by order, as the ones with the highest order are the hardest to place
    # note that this is actually reverse sorted, as the top order is -2-bottom order, but this 
    # is good, as we want to manipulate the list from the end
    ordered_keys = [k for k,v in reversed(sorted(orders_bot.items(), key=lambda o: o[1]))]
    def splits(keys):
        # distribute the points (keys) onto spaces
        if not keys:
            # done if we hit all components and all spaces are 0
            if all(distr) and all(s == 0 for s in space):
                legal_splits.append([[a for a in c] for c in distr])
            return
        else:
            # check if there are enough points left
            remaining_comp = len([hit for hit in distr if not hit]) # components that don't have a point yet
            if remaining_comp > len(keys):
                return
            current = keys.pop()
            current_order = -2 - orders_bot[current]
            # try to place current on all components
            for i in range(len(space)):
                if space[i] >= current_order:
                    space[i] -= current_order
                    distr[i].append(current)
                    splits(keys) # recursion
                    # undo changes:
                    space[i] += current_order
                    distr[i].pop()
            keys.append(current)
    # generate splits:
    splits(ordered_keys)
    return legal_splits

def _distribute_fully(points,n):
    ## iterator distributing list of points onto n components, where each component gets at least one point
    ## return a nested list of points
    # generate partitions into n subsets of permuted_points:
    for part in multiset_partitions(points, n):
        # we need to consider all permutations (multiset_partitions sorts!)
        for permuted_points in permutations(part):
            yield list(permuted_points) # permutations give tuples...

def _b_ary_gen(x,b,n):
    ## generator for reverse b-ary integers x of length n
    for _ in itertools.repeat(None,n):
        r = x % b
        x = (x-r)//b
        yield r

def _distribute_points(points,n):
    ## iterator distributing list of points onto n components (some might not receive any)
    ## return a nested list of points
    l = len(points)
    # n^l possibilities:
    for i in range(n**l):
        point_list = [[] for j in range(n)]
        # n-ary representation of i tells us where the points should go
        for pos,d in enumerate(_b_ary_gen(i,n,l)):
            point_list[d].append(points[pos])
        yield point_list

def _distribute_part_ordered(g,n):
    ## iterator of partitions (g_1,...,g_k) of g of length k <= n distributed onto n points
    ## return a list of length n (that sums to g)
    if n < g:
        maxi = n
    else:
        maxi = None
    for part_dict in partitions(g,m=maxi):
        # partitions are dictionaries {g_i : multiplicity}
        part = []
        for k in part_dict.keys():
            part += [k]*part_dict[k]
        # fill up with zeros:
        part += (n-len(part))*[0]
        yield part
        # we do not have to permute if everything else is permuted....
        # for perm_part in set(permutations(part)):
        #     yield perm_part

def isom_rep(L):
    """
    Return a list of representatives of isomorphism classes of L.

    TODO: optimise!
    """
    dist_list = []
    for g in L:
        if all(not g.is_isomorphic(h) for h in dist_list):
            dist_list.append(g)
    return dist_list

def comp_list(L,H):
    r"""
    Compare two lists of LevelGraphs (up to isomorphism).

    Returns a tuple: (list L without H, list H without L)
    """
    return ([g for g in L if not any(g.is_isomorphic(h) for h in H)],
            [g for g in H if not any(g.is_isomorphic(h) for h in L)])

def test_bic_algs(sig_list=None):
    """
    Compare output of bics and bic_alt.

    EXAMPLES ::

        sage: from admcycles.diffstrata import *
        sage: test_bic_algs()  # long time (45 seconds)  # skip, not really needed + long # doctest: +SKIP
        (1, 1): ([], [])
        (1, 1, 0, 0, -2): ([], [])
        (2, 0, -2): ([], [])
        (1, 0, 0, 1): ([], [])
        (1, -2, 2, 1): ([], [])
        (2, 2): ([], [])

        sage: test_bic_algs([(0,0),(2,1,1)])  # long time (50 seconds)  # doctest: +SKIP
        (0, 0): ([], [])
        (2, 1, 1): ([], [])

        sage: test_bic_algs([(1,0,-1),(2,),(4,),(1,-1)])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (1, 0, -1): ([], [])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (2,): ([], [])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (4,): ([], [])
        WARNING! This is not the normal BIC generation algorithm!
        Only use if you know what you're doing!
        (1, -1): ([], [])
    """
    if sig_list == None:
        sig_list = [(1,1),(1,1,0,0,-2),(2,0,-2),(1,0,0,1),(1,-2,2,1),(2,2)]
    for sig in sig_list:
        Sig = Signature(sig)
        print("%r: %r" % (Sig.sig, comp_list(bics(Sig.g,Sig.sig),bic_alt(Sig)))) 


from __future__ import absolute_import
from __future__ import print_function

from copy import deepcopy

# pylint does not know sage
from sage.structure.sage_object import SageObject # pylint: disable=import-error
from sage.modules.free_module import FreeModule # pylint: disable=import-error
from sage.rings.integer_ring import ZZ # pylint: disable=import-error
from sage.arith.functions import lcm # pylint: disable=import-error
from sage.misc.flatten import flatten # pylint: disable=import-error
from sage.functions.generalized import sign # pylint: disable=import-error
from sage.misc.cachefunc import cached_method # pylint: disable=import-error
from sage.graphs.graph import Graph # pylint: disable=import-error

from admcycles.admcycles import GraphIsom
from admcycles.stable_graph import StableGraph as stgraph
import admcycles.diffstrata.levelstratum
from admcycles.diffstrata.sig import Signature

def smooth_LG(sig):
    """
    The smooth (i.e. one vertex) LevelGraph in the stratum (sig).
    
    Args:
        sig (Signature): signature of the stratum.
    
    Returns:
        LevelGraph: The smooth graph in the stratum.
    """
    return LevelGraph.fromPOlist([sig.g],[list(range(1,sig.n+1))],[],sig.sig,[0])

class LevelGraph(SageObject):
    r"""
    Create a (stable) level graph.

    NOTE: This is a low-level class and should NEVER be invoced directly!
    Preferably, EmbeddedLevelGraphs should be used and these should be
    generated automatically by Stratum (or GeneralisedStratum).

    NOTE: We don't inherit from stgraph/StableGraph anymore, as LevelGraphs
    should be static objects!

    Extends admcycles stgraph to represent a level graph as a stgraph,
    i.e. a list of genera, a list of legs and a list of edges, plus a list of
    poleorders for each leg and a list of levels for each vertex.

    Note that the class will warn if the data is not admissible, i.e. if the
    graph is not stable or the pole orders at separating nodes across levels do
    not add up to -2 or -1 on the same level (unless this is suppressed with
    quiet=True).

    INPUT:

    genera : list
        List of genera of the vertices of length m.
    legs : list
        List of length m, where ith entry is list of legs attached to vertex i.
        By convention, legs are unique positive integers.
    edges : list
        List of edges of the graph. Each edge is a 2-tuple of legs.
    poleorders : dictionary
        Dictionary of the form leg number : poleorder
    levels : list
        List of length m, where the ith entry corresponds to the level of
        vertex i.
        By convention, top level is 0 and levels go down by 1, in particular
        they are NEGATIVE numbers!
    quiet : optional boolean (default = False)
        Suppresses most error messages.

    ALTERNATIVELY, the pole orders can be supplied as a list by calling
    LevelGraph.fromPOlist:
    poleorders : list
        List of order of the zero (+) or pole (-) at each leg. The ith element
        of the list corresponds to the order at the leg with the marking i+1
        (because lists start at 0 and the legs are positive integers).

    EXAMPLES ::

        Creating a level graph with three components on different levels of genus 1,
        3 and 0. The bottom level has a horizontal node.

        sage: from admcycles.diffstrata import *
        sage: G = LevelGraph.fromPOlist([1,3,0],[[1,2],[3,4,5],[6,7,8,9]],[(2,3),(5,6),(7,8)],[2,-2,0,6,-2,0,-1,-1,0],[-2,-1,0]); G  #  doctest: +SKIP
        LevelGraph([1, 3, 0],[[1, 2], [3, 4, 5], [6, 7, 8, 9]],[(3, 2), (6, 5), (7, 8)],{1: 2, 2: -2, 3: 0, 4: 6, 5: -2, 6: 0, 7: -1, 8: -1, 9: 0},[-2, -1, 0],True)

        or alternatively:

        sage: LevelGraph([1, 3, 0],[[1, 2], [3, 4, 5], [6, 7, 8, 9]],[(3, 2), (6, 5), (7, 8)],{1: 2, 2: -2, 3: 0, 4: 6, 5: -2, 6: 0, 7: -1, 8: -1, 9: 0},[-2, -1, 0],True)
        LevelGraph([1, 3, 0],[[1, 2], [3, 4, 5], [6, 7, 8, 9]],[(3, 2), (6, 5), (7, 8)],{1: 2, 2: -2, 3: 0, 4: 6, 5: -2, 6: 0, 7: -1, 8: -1, 9: 0},[-2, -1, 0],True)

        We get a warning if the graph has non-stable components: (not any more ;-))

        sage: G = LevelGraph.fromPOlist([1,3,0,0],[[1,2],[3,4,5],[6,7,8,9],[10]],[(2,3),(5,6),(7,8),(9,10)],[2,-2,0,6,-2,0,-1,-1,0,-2],[-3,-2,0,-1]); G  # doctest: +SKIP
        Warning: Component 3 is not stable: g = 0 but only 1 leg(s)!
        Warning: Graph not stable!
        LevelGraph([1, 3, 0, 0],[[1, 2], [3, 4, 5], [6, 7, 8, 9], [10]],[(3, 2), (6, 5), (7, 8), (9, 10)],{1: 2, 2: -2, 3: 0, 4: 6, 5: -2, 6: 0, 7: -1, 8: -1, 9: 0, 10: -2},[-3, -2, 0, -1],True)
    """
    def __init__(self,genera,legs,edges,poleorders,levels,quiet=False):
        checks = False
        if len(genera) != len(legs):
            raise ValueError('genera and legs must have the same length')
        self.genera = genera
        self.legs = legs
        self.edges = edges
        self.poleorders = poleorders
        self.levels = levels
        # the signature consists of the marked points that are not half-edges
        sig_list = tuple(poleorders[l] for l in self.list_markings())
        self.sig = Signature(sig_list)
        # associated stgraph...
        self._stgraph = None
        self.k = 1 # in case we want to implement k-differentials
        self._init_more() # initiate some things and make sure all is in order
        if checks:
            self.checkadmissible(quiet)
        self._has_long_edge = None
        self._is_long = {}
        self._is_bic = None

    @classmethod
    def fromPOlist(cls,genera,legs,edges,poleordersaslist,levels,quiet=False):
        """
        This gives a LevelGraph where the poleorders are given as a list, not
        directly as a dictionary.
        """
        # translate poleorder list to dictionary:
        sortedlegs = sorted(flatten(legs))
        if len(sortedlegs) != len(poleordersaslist):
             raise ValueError("Numbers of legs and pole orders do not agree!"+
                            " Legs: " + repr(len(sortedlegs)) +
                            " Pole orders: " + repr(len(poleordersaslist)))
        poleorderdict = { sortedlegs[i] : poleordersaslist[i] for i in range(len(sortedlegs)) }
        return cls(genera,legs,edges,poleorderdict,levels,quiet)

    def __repr__(self):
        return "LevelGraph("+self.input_as_string()+")"
    def __hash__(self):
        return hash(repr(self))
    def __str__(self):
        return ("LevelGraph " + repr(self.genera)+ ' ' + repr(self.legs) + ' '
                + repr(self.edges) + ' ' + repr(self.poleorders) + ' '
                + repr(self.levels))
    def input_as_string(self):
        """
        return a string that can be given as argument to __init__ to give self
        """
        return (repr(self.genera)+ ',' + repr(self.legs) + ','
                + repr(self.edges) + ',' + repr(self.poleorders) + ','
                + repr(self.levels)+',True')
    def __eq__(self,other):
        if not isinstance(other,LevelGraph):
            return False
        return (self.genera==other.genera) and (self.levels == other.levels) and (self.legs == other.legs) and (set(self.edges)==set(other.edges)) and (self.poleorders == other.poleorders)

    ## reimplementing the admcycles stuff we use:
    @cached_method
    def g(self):
        genus = sum(self.genera) + len(self.edges) - len(self.genera) + ZZ.one()
        assert genus == self.sig.g, "Signature %r does not match %r!" % (self.sig,self)
        return genus

    @property
    def stgraph(self):
        if self._stgraph is None:
            self._stgraph = stgraph([int(g) for g in self.genera],
                                    [[int(l) for l in leg_list] for leg_list in self.legs],
                                    self.edges)
        return self._stgraph

    @cached_method
    def list_markings(self,v=None):
        r"""
        Return the list of markings (non-edge legs) of self at vertex v.
        """
        if v is None:
            return tuple([j for v in range(len(self.genera)) for j in self.list_markings(v) ])
        s = set(self.legs[v])
        for e in self.edges:
            s -= set(e)
        return tuple(s)

    ##########################################################
    #### Interface:
    ####   --- provide a series of methods that translate various data
    ####   (maybe this can be optimised by smart caching later, so only
    ####    these should be used...)
    #### up to now, we have:
    ####    * vertex : (overloaded from admcycles)
    ####            INPUT : leg
    ####            returns : vertex of leg (index of genera)
    ####    * vertex_list :
    ####            returns : list of vertices (index of genera)
    ####    * edges_list :
    ####            returns : list of edges
    ####    * prongs_list :
    ####            returns : list of tuples (edge,prong)
    ####    * levelofvertex :
    ####            INPUT : index of genera
    ####            returns : level number (negative)
    ####    * levelofleg :
    ####            INPUT : leg label (given as positive integer)
    ####            returns : level number (negative)
    ####    * orderatleg :
    ####            INPUT : leg label (given as positive integer)
    ####            returns : poleorder (integer)
    ####    * ordersonvertex :
    ####            INPUT : index of genera
    ####            returns : list of poleorders (integers)
    ####    * verticesonlevel :
    ####            INPUT : level number (negative)
    ####            returns : list of indices of genera
    ####    * edgesatvertex :
    ####            INPUT : index of genera
    ####            returns : list of edges
    ####    * legsatvertex :
    ####            INPUT : index of genera
    ####            returns : list of legs
    ####    * simplepolesatvertex :
    ####            INPUT : index of genera
    ####            returns : list of legs with pole order -1
    ####    * genus :
    ####            INPUT : index of genera
    ####            returns : genus
    ####    * codim :
    ####            returns : codimension
    ####    * is_bic :
    ####            returns : boolean
    ####    * edgesatlevel :
    ####            INPUT : level number (negative)
    ####            returns : list of edges with at least one node at that level (or horizontal)
    ####    * horizontaledgesatlevel :
    ####            INPUT : level number (negative)
    ####            returns : list of horizontal edges
    ####    * prong :
    ####            INPUT : edge
    ####            returns : prong order at edge
    ####    * nextlowerlevel :
    ####            INPUT : level number
    ####            returns : level number of next lower level
    ####    * lowest_level :
    ####            returns : level number of lowest level
    ####    * is_horizontal :
    ####            INPUT : edge (or none for whole graph)
    ####            returns : boolean success
    ####    * has_loop :
    ####            INPUT : vertex
    ####            returns : boolean success
    ####    * edgesgoingupfromlevel :
    ####            INPUT : level
    ####            returns : list of edges with e[1] on level
    ####    * verticesabove :
    ####            INPUT : level
    ####            returns : list of vertices with level > l
    ####    * edgesabove :
    ####            INPUT : level
    ####            returns : list of edges with level of each vertex > l
    ####    * crosseslevel :
    ####            INPUT : edge, level
    ####            returns : boolean
    ####    * edgesgoingpastlevel :
    ####            INPUT : level
    ####            returns : list of edges with start level > l and
    ####                      end level < l
    ####    * _pointtype :
    ####            INPUT : order (integer)
    ####            returns : pole/marked point/zero (string)
    ####    * is_meromorphic :
    ####            returns : boolean (check for pole among marked points)
    ####    * highestorderpole :
    ####            INPUT : vertex
    ####            returns : leg name (at v) or -1 if none found
    #########################################################
    @cached_method
    def vertex(self,leg):
        """
        The vertex (as index of genera) where leg is located.

        Args:
            leg (int): leg

        Returns:
            int: index of genera
        """
        return self.legdict[leg]
    @cached_method
    def vertex_list(self):
        return list(range(len(self.genera)))
    @cached_method
    def edges_list(self):
        return self.edges
    @cached_method
    def prongs_list(self):
        return [(e,p) for e,p in self.prongs.items()]
    @cached_method
    def levelofvertex(self,v):
        return self.levels[v]
    @cached_method
    def levelofleg(self,leg):
        return self.levelofvertex(self.vertex(leg))
    @cached_method
    def orderatleg(self,leg):
        return self.poleorders[leg]
    @cached_method
    def ordersonvertex(self,v):
        return [self.orderatleg(leg) for leg in self.legsatvertex(v)]
    @cached_method
    def verticesonlevel(self,level):
        return [v for v in range(len(self.genera)) if self.levelofvertex(v) == level]
    @cached_method
    def edgesatvertex(self,v):
        return [e for e in self.edges if self.vertex(e[0]) == v or
                                        self.vertex(e[1]) == v]
    @cached_method
    def legsatvertex(self,v):
        return self.legs[v]
    @cached_method
    def is_halfedge(self,leg):
        """
        Check if leg has an edge attached to it.
        """
        for e in self.edges:
            if e[0] == leg or e[1] == leg:
                return True
        return False
    @cached_method
    def simplepolesatvertex(self,v):
        return [l for l in self.legsatvertex(v) if self.orderatleg(l) == -1]
    @cached_method
    def genus(self,v=None):
        """
        Return the genus of vertex v.
        If v==None, return genus of the complete LevelGraph.
        """
        if v == None:
            return self.g() # from admcycles
        else:
            return self.genera[v]
    @cached_method
    def numberoflevels(self):
        return len(set(self.levels))
    def is_bic(self):
        if self._is_bic is None:
            self._is_bic = self.numberoflevels() == 2 and not self.is_horizontal()
        return self._is_bic
    @cached_method
    def codim(self):
        """
        Return the codim = No. of levels of self + horizontal edges.
        """
        return (self.numberoflevels() - 1
                    + len([e for e in self.edges if self.is_horizontal(e)]))
    @cached_method
    def edgesatlevel(self,level):
        """
        Return list of edges with at least one node at level.
        """
        return [e for e in self.edges if self.levelofleg(e[0]) == level or
                                        self.levelofleg(e[1]) == level]
    @cached_method
    def horizontaledgesatlevel(self,level):
        return [e for e in self.edgesatlevel(level) if self.is_horizontal(e)]
    @cached_method
    def prong(self,e):
        """
        The prong order is the pole order of the higher-level pole +1
        """
        if self.levelofleg(e[0]) > self.levelofleg(e[1]):
            return self.orderatleg(e[0]) + 1
        else:
            return self.orderatleg(e[1]) + 1
    @cached_method
    def nextlowerlevel(self,l):
        """
        Return the next lower level number or  False if no legal level
        (e.g. lowest level) is given as input.

        Point of discussion: Is it better to tidy up the level numbers
        to be directly ascending or not?

        Pro tidy: * this becomes obsolete ;)
                  * easier to check isomorphisms?
        Con tidy: * we "see" where we came from
                  * easier to undo/glue back in after cutting
        """
        try:
            llindex = self.sortedlevels.index(l) - 1
        except ValueError:
            return False
        if llindex == -1:
            return False
        else:
            return self.sortedlevels[llindex]
    @cached_method
    def internal_level_number(self,i):
        """
        Return the internal i-th level, e.g.

        self.levels = [0,-2,-5,-3]

        then

        internal_level_number(0) is 0
        internal_level_number(1) is -2
        internal_level_number(2) is -3
        internal_level_number(3) is -4

        Returns None if the level does not exist.
        """
        reference_levels = list(reversed(sorted(list(set(self.levels)))))
        i = abs(i)
        if i >= len(reference_levels):
            return None
        else:
            return reference_levels[i]
    @cached_method
    def level_number(self,l):
        """
        Return the relative level number (positive!) of l, e.g.

        self.levels = [0,-2,-5,-3]

        then

        level_number(0) is 0
        level_number(-2) is 1
        level_number(-3) is 2
        level_number(-5) is 3

        Returns None if the level does not exist.
        """
        reference_levels = list(reversed(sorted(list(set(self.levels)))))
        try:
            return reference_levels.index(l)
        except ValueError:
            return None
    @cached_method
    def is_long(self,e):
        """
        Check if edge e is long, i.e. passes through more than one level.
        """
        try:
            return self._is_long[e]
        except KeyError:
            il = abs(self.level_number(self.levelofleg(e[0])) - self.level_number(self.levelofleg(e[1]))) > 1
            self._is_long[e] = il
            return il
    @cached_method
    def has_long_edge(self):
        if self._has_long_edge is None:
            for e in self.edges:
                if self.is_long(e):
                    self._has_long_edge = True
                    break
            else:
                self._has_long_edge = False
        return self._has_long_edge
    @cached_method
    def lowest_level(self):
        return self.sortedlevels[0]
    @cached_method
    def is_horizontal(self,e=None):
        """
        Check if edge e is a horizontal edge or if self has a horizontal edge.
        """
        if e == None:
            for e in self.edges:
                if self.is_horizontal(e):
                    return True
            return False
        if not (e in self.edges):
            print("Warning: " + repr(e) + " is not an edge of this graph!")
            return False
        if self.levelofleg(e[0]) == self.levelofleg(e[1]):
            return True
        else:
            return False
    @cached_method
    def has_loop(self,v):
        """
        Check if vertex v has a loop.
        """
        for e in self.edgesatvertex(v):
            if self.vertex(e[0]) == self.vertex(e[1]):
                return True
        return False
    @cached_method
    def edgesgoingupfromlevel(self,l):
        """
        Return the edges going up from level l.

        This uses that e[0] is not on a lower level than e[1]!
        """
        return [e for e in self.edges if self.levelofleg(e[1])==l and
                                        not self.is_horizontal(e)]
    @cached_method
    def verticesabove(self,l):
        """
        Return list of all vertices above level l.

        If l is None, return all vertices.
        """
        if l is None:
            return list(range(len(self.genera)))
        return [v for v in range(len(self.genera)) if self.levelofvertex(v) > l]
    @cached_method
    def edgesabove(self,l):
        """
        Return a list of all edges above level l, i.e. start and end vertex
        have level > l.
        """
        return [e for e in self.edges if self.levelofleg(e[0]) > l
                                        and self.levelofleg(e[1]) > l]
    @cached_method
    def crosseslevel(self,e,l):
        """
        Check if e crosses level l (i.e. starts > l and ends <= l)
        """
        return self.levelofleg(e[0]) > l and self.levelofleg(e[1]) <= l
    @cached_method
    def edgesgoingpastlevel(self,l):
        """
        Return list of edges that go from above level l to below level l.
        """
        return [e for e in self.edges if self.levelofleg(e[0]) > l
                                        and self.levelofleg(e[1]) < l]
    @cached_method
    def _pointtype(self,order):
        if order < 0:
            return "pole"
        elif order == 0:
            return "marked point"
        else:
            return "zero"
    @cached_method
    def is_meromorphic(self):
        """
        Returns True iff at least one of the MARKED POINTS is a pole.
        """
        return any(self.orderatleg(l) < 0 for l in self.list_markings())
    @cached_method
    def highestorderpole(self,v):
        """
        Returns the leg with the highest order free pole at v, -1 if no poles found.
        """
        minorder = 0
        leg = -1
        for l in self.list_markings(v):
            if self.orderatleg(l) < minorder:
                minorder = self.orderatleg(l)
                leg = l
        return leg
    #########################################################
    #### end interface
    #########################################################

    #########################################################
    #### Checks
    #########################################################
    @cached_method
    def checkadmissible(self,quiet=False):
        """
        Run all kinds of checks on the level graph. Currently:
         * Check if orders on each komponent add up to k*(2g-2)
         * Check if graph is stable
         * Check if orders at edges sum to -2
         * Check if orders respect level crossing
        """
        admissible = True
        if not self.checkorders(quiet):
            if not quiet:
                print("Warning: Illegal orders!")
            admissible = False
        if not self.is_stable(quiet):
            if not quiet:
                print("Warning: Graph not stable!")
            admissible = False
        if not self.checkedgeorders(quiet):
            if not quiet:
                print("Warning: Illegal orders at a node!")
            admissible = False
        return admissible

    @cached_method
    def checkorders(self,quiet=False):
        """
        Check if the orders add up to k*(2g-2) on each component.
        """
        for v, g in enumerate(self.genera):
            if sum(self.ordersonvertex(v)) != self.k*(2*g-2):
                if not quiet:
                    print("Warning: Component " + repr(v) + " orders add up to " +
                        repr(sum(self.ordersonvertex(v))) +
                        " but component is of genus " + repr(g))
                return False
        return True

    @cached_method
    def is_stable(self,quiet=False):
        """
        Check if graph is stable.
        """
        # total graph:
        e = 2*self.genus() - 2 + len(self.leglist)
        if e < 0:
            if not quiet:
                print("Warning: Total graph not stable! 2g-2+n = " + repr(e))
            return False

        # components:
        for v in range(len(self.genera)):
            if 3*self.genus(v) - 3 + len(self.legsatvertex(v)) < 0:
                if not quiet:
                    print("Warning: Component " + repr(v) + " is not stable: g = "
                        + repr(self.genus(v)) + " but only "
                        + repr(len(self.legsatvertex(v))) + " leg(s)!")
                return False
        return True

    @cached_method
    def checkedgeorders(self,quiet=False):
        """
        Check that the orders at nodes (i.e. edges) sum to -1 and that lower
        level has lower order.
        """
        for e in self.edges:
            orders = self.orderatleg(e[0]) + self.orderatleg(e[1])
            if orders != -2:
                if not quiet:
                    print("Warning: Orders at edge " + repr(e) + " add up to " +
                        repr(orders) + " instead of -2!")
                return False
            # iff the pole order at e[0] is > the poleorder at e[1], e[0] should be on a higher level than e[1]
            if (sign(self.orderatleg(e[0]) - self.orderatleg(e[1]))
                    != sign(self.levelofleg(e[0]) - self.levelofleg(e[1]))):
                if not quiet:
                    print("Warning: Orders at edge " + repr(e) +
                        " do not respect level crossing!")
                    print("Poleorders are", 
                        (self.orderatleg(e[0]),self.orderatleg(e[1])), 
                        "but levels are",  
                        (self.levelofleg(e[0]),self.levelofleg(e[1]))
                    )
                return False
        return True

    #################################################################
    ### End checks
    #################################################################

    #################################################################
    ### Cleanup functions
    ### USE ONLY IN __init__!!! LevelGraphs should be static!!!
    #################################################################
    def _gen_prongs(self):
        """
        Generate the dictionary edge : prong order.
        The prong order is the pole order of the higher-level pole +1
        """
        return { e : self.prong(e) for e in self.edges }

    def _init_more(self):
        """
        This should be used _only_ for intialisation!

        Make sure everything is in order, in particular:
        * sortedlevels is fine
        * leglist is fine
        * legdict is fine
        * maxleg is fine
        * maxlevel is fine
        * prongs are fine
        * underlying graph is fine
        """
        self.sortedlevels = sorted(self.levels)
        # legs is a nested list:
        self.leglist = flatten(self.legs)
        # legs as dictionary
        self.legdict = { l:v for v in range(len(self.genera))
                             for l in self.legs[v] }
        self.maxleg = max([max(j+[0]) for j in self.legs])
        # maxlevel is named a bit misleading, as it is the LOWEST level
        # (the max of the absolute values...)
        self.maxlevel = max([abs(l) for l in self.levels])
        # the prong orders are stored in a dictionary edge : prong order
        self.prongs = self._gen_prongs()
        # construct the "underlying graph" (as a sage graph)
        # the edges are labeled by their "real" name,
        # the vertices are of the form (i,g_i,'LG')
        # NOTE: In the case of an EmbeddedLevelGraph, vertices "at infinity" might
        # be added to this!
        self.underlying_graph = Graph([[self.UG_vertex(i) for i in range(len(self.genera))],
                    [(self.UG_vertex(self.vertex(e[0])), self.UG_vertex(self.vertex(e[1])), e)
                                        for e in self.edges]],
                    format='vertices_and_edges', loops=True, multiedges=True)

    def UG_vertex(self, v):
        """
        Vertex of the underlying graph corresponding to v.
        
        Args:
            v (int): vertex number (index of self.genera)
        
        Returns:
            tuple: tuple encoding the vertex of the Underlying Graph.
        """
        return (v, self.genera[v], 'LG')

    #################################################################
    #### Extract a subgraph
    #################################################################
    def extract(self,vertices,edges):
        """
        Extract the subgraph of self (as a LevelGraph) consisting of vertices
        (as list of indices) and edges (as list of edges).

        Returns the levelgraph consisting of vertices, edges and all legs on
        vertices (of self) with their original poleorders and the original
        level structure.
        """
        newvertices = deepcopy([self.genera[i] for i in vertices])
        newlegs = deepcopy([self.legs[i] for i in vertices])
        newedges = deepcopy(edges)
        newpoleorders = deepcopy({l : self.orderatleg(l)
                                    for l in flatten(newlegs)})
        newlevels = deepcopy([self.levels[i] for i in vertices])
        return LevelGraph(newvertices,newlegs,newedges,newpoleorders,newlevels)

    def levelGraph_from_subgraph(self,G):
        """
        Returns the LevelGraph associated to a subgraph of underlying_graph
        (with the level structure induced by self)
        """
        return self.extract(G.vertices(),G.edge_labels())

    def stratum_from_level(self,l,excluded_poles=None):
        """
        Return the LevelStratum at (relative) level l.

        Args:
            l (int): relative level number (0,...,codim)
            excluded_poles (tuple, defaults to None): a list of poles (legs of the graph,
                that are marked points of the stratum!) to be ignored for r-GRC.

        Returns:
            LevelStratum: the LevelStratum, i.e. 
                * a list of Signatures (one for each vertex on the level)
                * a list of residue conditions, i.e. a list [res_1,...,res_r]
                    where each res_k is a list of tuples [(i_1,j_1),...,(i_n,j_n)]
                    where each tuple (i,j) refers to the point j (i.e. index) on the
                    component i and such that the residues at these points add up
                    to 0.
                * a dictionary of legs, i.e. n -> (i,j) where n is the original
                    number of the point (on the LevelGraph self) and i is the 
                    number of the component, j the index of the point in the signature tuple.
                Note that LevelStratum is a GeneralisedStratum together with
                a leg dictionary.
        """
        if self.is_horizontal():
            print("Error: Cannot extract levels of a graph with horizontal edges!")
            return None
        internal_l = self.internal_level_number(l)
        # first, we extract the obvious information from self at this level:
        vv = self.verticesonlevel(internal_l)
        legs_on_level = []
        sigs = []
        res_cond = []
        leg_dict = {}
        for i,v in enumerate(vv):
            legs_on_level += [deepcopy(self.legsatvertex(v))]
            leg_dict.update([(n,(i,j)) for j,n in enumerate(legs_on_level[-1])])
            sig_v = Signature(tuple(self.orderatleg(leg) for leg in legs_on_level[-1]))
            sigs += [sig_v]
        # Now we hunt for residue conditions.
        # We work with the underlying graph to include the "level at infinity":
        # We "cheat" here, to catch the edges going up from level l: These are the
        # edges of the graph above level l-1 that have a vertex on level l:
        # (see below for parsing of these edges)
        ee = [] 
        for e in self.UG_above_level(internal_l - 1).edges():
            for UG_vertex in e[:2]:
                if UG_vertex[2] == 'res':
                    continue
                if self.levelofvertex(UG_vertex[0]) == internal_l:
                    ee.append(e)
                    break
        G = self.UG_above_level(internal_l)
        conn_comp = G.connected_components_subgraphs()
        # Residue conditions are obtained by sorting the poles on this level
        # by connected component of G they are attached to (note that this
        # includes the "level at infinity"!)
        for c in conn_comp:
            # if there is a free pole in c, we ignore this component
            if self._redeemed_by_merom_in_comp(c,excluded_poles=excluded_poles):
                continue
            res_cond_c = []
            # Otherwise we record all poles to attached to this component in one
            # residue condition:
            for e in ee:
                # We work with edges of the underlying graph here, so we have to
                # be a bit more careful. Recall that edges of the underlying graph are of
                # the form:
                # * (UG_vertex, UG_vertex, edge name) if it's an edge of LG or
                # * (UG_vertex, UG_vertex, resiedgej) if it's an edge to level infinity
                # Here UG_vertex is either of the form (vertex number, genus, 'LG') or 
                # (i, 0, 'res') where
                # i is the number of the vertex at infinity and resiedgej is the edge
                # connecting the residue i to the leg j (of self).
                if e[0] in c or e[1] in c:
                    # res_cond contains tuples (i,j) where i is the index of the 
                    # component and j the index of the pole in sig_i.
                    #
                    # We need find the vertex on this level (i.e. not in c)
                    if e[0] in c:
                        UG_vertex = e[1]
                    else:
                        UG_vertex = e[0]
                    # to find the pole on v, we have to distinguish if this is an
                    # edge of self or if it connects to a residue at infinity:
                    edge = e[2]
                    if 'res' in edge:
                        # the leg is the number after 'edge' in the edge string:
                        _, leg = edge.split('edge')
                        leg = int(leg)
                    else:
                        # edge consists of (leg_top,leg_bot) and leg_bot is on level i:
                        leg = edge[1]
                    # We retrieve the stratum point from leg_dict:
                    res_cond_c += [leg_dict[leg]]
            if res_cond_c:
                res_cond += [res_cond_c]
        return admcycles.diffstrata.levelstratum.LevelStratum(sigs,res_cond,leg_dict)

    def UG_vertices_above_level(self,l):
        """
        The vertices above (internal) level l (including possible vertices at infinity).

        Used for checking residue conditions.
        
        Args:
            l (int): internal level number
        
        Returns:
            list: list of vertices of self.underlying_graph
        """
        # note that verticesabove checks for >l
        vertices_above = [self.UG_vertex(i) for i in self.verticesabove(l)]
        vertices_above += [v for v in self.underlying_graph.vertices() if v[2] == 'res']
        return vertices_above
    
    def UG_above_level(self,l):
        """
        Subgraph of Underlying Graph (including vertices at infinity) (strictly) above
        (relative!) level l.
        
        Args:
            l (int): relative level number
        
        Returns:
            SAGE Graph: subgraph of self.underlying_graph with all vertices strictly
                above level l.
        """
        internal_l = self.internal_level_number(l)
        vertices = self.UG_vertices_above_level(internal_l)
        abovegraph = self.underlying_graph.subgraph(vertices)
        return abovegraph

    #################################################################
    #### Check Residue condition (via inconvenient vertices/edges)
    #################################################################

    @cached_method
    def is_inconvenient_vertex(self,v):
        """
        Check if vertex is inconvenient, i.e.
        * g = 0
        * no simple poles
        * there exists a pole order m_i such that m_i > sum(m_j)-p-1
        Return boolean
        """
        # inconvenient vertices are of genus 0 and have no simple poles.
        if self.genus(v) > 0 or len(self.simplepolesatvertex(v)) > 0:
            return False
        # check inequality
        ll = self.legsatvertex(v)
        poles = [l for l in ll if self.orderatleg(l) < 0]
        polesum = sum([-self.orderatleg(l) for l in poles])
        p = len(poles)
        for l in ll:
            if self.orderatleg(l) > polesum - p - 1:
                return True
        return False

    def _redeemed_by_merom_in_comp(self,G,excluded_poles=None):
        """
        Check if there is a pole in the subgraph G (intended to be a connected
        component above a vertex).

        excluded_poles are ignored (for r-GRC).

        Returns boolean (True = exists a pole)
        """
        if excluded_poles is None:
            excluded_poles = []
        for w in G.vertices():
            if w[2] == 'res':  # vertex at infinity
                continue
            for l in self.list_markings(w[0]):
                if (self.orderatleg(l) < 0) and (not l in excluded_poles):
                    return True
        return False

    @cached_method
    def is_illegal_vertex(self,v,excluded_poles=None):
        """
        Check if vertex is inconvenient and is not redeemed, i.e.
        * v is inconvenient
        * there are no two seperate edges to the same connected component above (i.e. loop above)
        * if meromorphic: v is not connected to higher level marked poles

        We can also pass a tuple (hashing!) of poles that are to be excluded because of 
        residue conditions.

        Return boolean
        """
        if not self.is_inconvenient_vertex(v):
            return False
        l = self.levelofvertex(v)
        # in the underlying_graph, vertices are of the form (index in genera, genus, 'LG'):
        v_graph = self.UG_vertex(v)
        # edges not going down from v:
        ee = [e for e in self.edgesatvertex(v) if self.levelofleg(e[0]) >= l
                                                and self.levelofleg(e[1]) >= l]
        # in the holomorphic case, legal inconvenient vertices need at least two edges not going down
        if len(ee) < 1 and not self.is_meromorphic():
            return True
        # check if v has two edges into one connected component that don't lie below v,
        # i.e. if v can be redeemed:
        abovegraph = self.UG_above_level(l-1)
        cc = self.underlying_graph.subgraph([w
                for w in abovegraph.connected_component_containing_vertex(v_graph)
                        if w[2] == 'res' or self.levelofvertex(w[0]) >= l])
        cc.delete_vertex(v_graph)
        conn_comp = cc.connected_components_subgraphs()
        if excluded_poles is None:
            excluded_poles = []
        freepoles = len([l for l in self.list_markings(v) 
                            if (self.orderatleg(l) < 0) and (not l in excluded_poles)])
        for G in conn_comp:
            # edges from v to G:
            # (Careful: cc does not have any edges with v anymore, so we must use abovegraph!)
            eeG = [e for e in abovegraph.edges() 
                                if (e[0] == v_graph and e[1] in G.vertices())
                                            or (e[1] == v_graph and e[0] in G.vertices())]
            if len(eeG) > 1:
                # redeemed :-)
                return False
            # in the meromorphic case, we also check if a "free" pole exists in a connected component above v
            if self.is_meromorphic() and self._redeemed_by_merom_in_comp(G,excluded_poles=excluded_poles):
                freepoles += 1
        if freepoles >= 2:
            # redeemed :-)
            return False
        return True

    @cached_method
    def is_illegal_edge(self,e,excluded_poles=None):
        """
        Check if edge is illegal, i.e. if
        * e is horizontal (not loop) and
        * there no simple loop over e
        * there is no "free" pole over each end point

        excluded_poles may be a tuple (hashing!) of poles to be excluded for r-GRC.

        Return boolean
        """
        if not self.is_horizontal(e) or (e[0] == e[1]):
            return False
        # check if there is a simple loop above e, i.e. if the subgraph
        # above e is still connected after e is removed (e is not a bridge)
        l = self.levelofleg(e[0])
        # note that verticesabove checks for >l
        abovegraph = self.UG_above_level(l-1)
        # edges are encoded by (self.vertex(e[0]),self.vertex(e[1]),e) in
        # underlying_graph!
        cut = abovegraph.is_cut_edge((self.UG_vertex(self.vertex(e[0])), self.UG_vertex(self.vertex(e[1])), e))
        # True = e is a cut-edge, i.e. connected components increase when removing e, i.e. no loop above, 
        #       i.e. illegal unless also meromorphic
        # False = graph has a loop, i.e. not illegal
        if not cut:
            return False
        else:
            if self.is_meromorphic():
                # if there are "free" poles above the endpoints, we are still fine
                abovegraph.delete_edge((self.UG_vertex(self.vertex(e[0])), self.UG_vertex(self.vertex(e[1])), e))
                polesfound = 0
                for l in e:
                    v = self.vertex(l)
                    G = abovegraph.connected_component_containing_vertex(self.UG_vertex(v))
                    if self._redeemed_by_merom_in_comp(abovegraph.subgraph(G),excluded_poles=excluded_poles):
                        polesfound += 1
                if polesfound == 2:
                    # we are fine
                    return False
        # no redemption :/
        return True

    @cached_method
    def is_legal(self,quiet=False,excluded_poles=None):
        """
        Check if self has illegal vertices or edges.

        excluded_poles may be a tuple (hashing!) of poles to be excluded for r-GRC.

        Return boolean
        """
        for v in range(len(self.genera)):
            if self.is_illegal_vertex(v,excluded_poles=excluded_poles):
                if not quiet:
                    print("Vertex", v, "is illegal!")
                return False
        for e in self.edges:
            if self.is_illegal_edge(e,excluded_poles=excluded_poles):
                if not quiet:
                    print("Edge", e, "is illegal!")
                return False
        return True

    #################################################################
    #### Squishing ---- i.e. contracting horizontal edges / levels
    #################################################################

    def squish_horizontal(self,e,adddata=False):
        """
        Squish the horizontal edge e and return the new graph.
        If adddata=True is specified, we additionally return the level graph
        that was removed together with the data needed to glue it back in,
        i.e. such that composition with split_horizontal is the identity.

        More precisely: return a tuple (G,(v,dl,lg,loop)) consisting of
         * the contracted LevelGraph G
         * the "remaining" vertex v
         * a dictionary dl remembering how the legs were split between the
            original vertices
         * a list lg consisting of the genera of the original vertices
         * a boolean loop
        """
        # make sure v <= w
        [v,w] = sorted([self.vertex(e[0]),self.vertex(e[1])])

        # generate extra data
        if adddata:
            if v == w:
                loop = True
                lg = [self.genus(v)]
            else:
                loop = False
                lg = [self.genus(v),self.genus(w)]
            dl = dict([(l,0) for l in self.legsatvertex(v)] +
                      [(l,1) for l in self.legsatvertex(w)])
            del dl[e[0]]
            del dl[e[1]]
            extradata = (v,dl,lg,loop)

        if not self.is_horizontal(e):
            print("Warning: " + repr(e) + " is not a horizontal edge -- Not contracting.")
            if adddata:
                return (self,extradata)
            else:
                return self

        newgenera = deepcopy(self.genera)
        newlegs = deepcopy(self.legs)
        newedges = deepcopy(self.edges)
        newpoleorders = deepcopy(self.poleorders)
        newlevels = deepcopy(self.levels)
        if v == w:
            # --horizontal loop getting contracted---
            # add genus to node:
            newgenera[v] += 1
        else:
            # --horizontal edge getting contracted---
            # transfer genus and legs from w to v:
            newgenera[v] += newgenera[w]
            newlegs[v] += newlegs[w]
            # remove all traces of w
            newgenera.pop(w)
            newlegs.pop(w)
            newlevels.pop(w)
        # remove edge:
        newedges.remove(e)
        # remove legs
        newlegs[v].remove(e[0])
        newlegs[v].remove(e[1])
        # remove poleorders
        del newpoleorders[e[1]]
        del newpoleorders[e[0]]

        returngraph = LevelGraph(newgenera,newlegs,newedges,newpoleorders,newlevels,True)

        if adddata:
            return (returngraph,extradata)
        else:
            return returngraph

    def squish_vertical_slow(self,l,adddata=False,quiet=False):
        """
        Squish the level l (and the next lower level!) and return the new graph.
        If addata=True is specified, we additionally return a boolean that tells
        us if a level was squished and the legs that no longer exist.

        More precisely, adddata=True returns a tuple (G,boolean,legs) where
        * G is the (possibly non-contracted) graph and
        * boolean tells us if a level was squished
        * legs is the (possibly empty) list of legs that don't exist anymore

        Implementation:
        At the moment, we remember all the edges (ee) that connect level l to
        level l-1. We then create a new LevelGraph with the same data as self,
        the only difference being that all vertices of level l-1 have now been
        assigned level l. We then remmove all (now horizontal!) edges in ee.

        In particular, all points and edges retain their numbering (only the level might have changed).

        WARNING: Level l-1 does not exist anymore afterwards!!!

        Downside: At the moment we get a warning for each edge in ee, as these
        don't have legal pole orders for being horizontal (so checkedgeorders
        complains each time the constructor is called :/).
        """
        if l == None:
            return self
        ll = self.nextlowerlevel(l)
        if ll == False:
            if not quiet:
                print("Warning: Illegal level to contract: " + repr(l))
            if adddata:
                return (self,False,None)
            else:
                return self

        if not quiet:
            print("Squishing levels", l, "and", ll)

        vv = self.verticesonlevel(ll)
        # edges that go to next lower level, i.e. will be contracted
        ee = [e for e in self.edgesatlevel(l) if self.levelofleg(e[1]) == ll
                                                and not self.is_horizontal(e)]

        if not quiet:
            print("Contracting edges", ee)

        newgenera = deepcopy(self.genera)
        newlegs = deepcopy(self.legs)
        newedges = deepcopy(self.edges)
        newpoleorders = deepcopy(self.poleorders)
        newlevels = deepcopy(self.levels)

        # raise level
        for v in vv:
            newlevels[v] = l
        returngraph = LevelGraph(newgenera,newlegs,newedges,newpoleorders,newlevels,True)
        # contract edges that are now horizontal:
        for e in ee:
            returngraph = returngraph.squish_horizontal(e)

        if adddata:
            return (returngraph,True,flatten(ee))
        else:
            return returngraph

    def squish_vertical(self,l,adddata=False,quiet=True):
        """
        Squish the level l (and the next lower level!) and return the new graph.

        Implementation:
        At the moment, we remember all the edges (ee) that connect level l to
        level l-1. We then create a new LevelGraph with the same data as self,
        the only difference being that all vertices of level l-1 have now been
        assigned level l. We then remmove all edges in ee.

        (In contrast to the old squish_vertical_slow, this is now done in
        one step and not recursively so we avoid a lot of the copying of graphs.)

        In particular, all points and edges retain their numbering (only the level might have changed).

        WARNING: Level l-1 does not exist anymore afterwards!!!
        
        Args:
            l (int): (internal) level number
            adddata (bool, optional): For legacy reasons, do not use. Defaults to False.
            quiet (bool, optional): No output. Defaults to True.
        
        Returns:
            LevelGraph: new levelgraph with one level less.
        """
        if l == None:
            return self
        ll = self.nextlowerlevel(l)
        levelset = set([l,ll])
        if ll == False:
            if not quiet:
                print("Warning: Illegal level to contract: " + repr(l))
            else:
                return self
        vertices = self.verticesonlevel(l) + self.verticesonlevel(ll)
        # edges that go to next lower level, i.e. will be contracted
        edges = [e for e in self.edges 
                    if set([self.levelofleg(e[0]), self.levelofleg(e[1])]) == levelset]
        # we use two dictionaries for quick lookup while we reorder the info:
        genus_of_vertex = {v : self.genus(v) for v in vertices}
        vertex_of_leg = {leg : v for v in vertices for leg in self.legsatvertex(v)}
        legs_of_vertex = {v : self.legsatvertex(v)[:] for v in vertices}
        deleted_legs = []
        while edges:
            start, end = edges.pop()
            v = vertex_of_leg[start]
            w = vertex_of_leg[end]
            del vertex_of_leg[start]
            del vertex_of_leg[end]
            legs_of_vertex[v].remove(start)
            legs_of_vertex[w].remove(end)
            deleted_legs.extend([start,end])
            if v == w:
                # if e (now) connects a vertex to itself, increase its genus
                genus_of_vertex[v] += 1
            else:
                # if e connects two different vertices, combine their data
                # (moving everything from w to v)
                genus_of_vertex[v] += genus_of_vertex[w]
                del genus_of_vertex[w]
                for leg in legs_of_vertex[w]:
                    vertex_of_leg[leg] = v
                    legs_of_vertex[v].append(leg)
                del legs_of_vertex[w]
        # Build new graph with this data:
        # We use sets for more efficient lookup:
        vertices = set(vertices)
        deleted_legs = set(deleted_legs)

        newgenera = []
        newlegs = []
        newlevels = []
        # we copy the untouched vertices and insert the new ones:
        for v in range(len(self.genera)):
            if v in vertices:
                if v in genus_of_vertex:
                    newgenera.append(genus_of_vertex[v])
                    newlegs.append(legs_of_vertex[v][:])
                    newlevels.append(l)
                # otherwise it has been deleted
            else:
                newgenera.append(self.genus(v))
                newlegs.append(self.legsatvertex(v)[:])
                newlevels.append(self.levelofvertex(v))

        # for the edges, we remove those with a delted edge:
        newedges = []
        for start, end in self.edges:
            if start in deleted_legs:
                assert end in deleted_legs
                continue
            assert not end in deleted_legs
            newedges.append((start, end))
        # for the new poleorders, we only have to remove the deleted legs:
        newpoleorders = {leg : p for leg, p in self.poleorders.items() if not leg in deleted_legs}
        
        return LevelGraph(newgenera,newlegs,newedges,newpoleorders,newlevels,True)

    def delta(self,k,quiet=False):
        """
        Squish all levels except for the k-th.

        Note that delta(1) contracts everything except top-level and that 
        the argument is interpreted via internal_level_number

        WARNING: Currently giving an out of range level (e.g. 
            0 or >= maxlevel) squishes the entire graph!!!

        Return the corresponding divisor.
        """
        G = self
        if self.is_horizontal():
            print("Error: Cannot delta a graph with horizontal edges!")
            return deepcopy(G)
        # recursive squishing forces us to go backwards, as the lower squished
        # level is always removed.
        for i in reversed(range(self.numberoflevels()-1)):
            # unfortunately delta and squish use slightly incompatible level
            # numbering, as the "illegal" squish here is k-1
            if abs(k)-1 == i:
                continue
            if not quiet:
                print("Squishing level", i)
            # note that it is important that internal_level_number of
            # self and not G is called in the argument, as the relative
            # numbering of levels in G changes, but the internal names don't
            G = G.squish_vertical(self.internal_level_number(i),quiet=quiet)
        return G

    #################################################################
    #### end Squishing
    #################################################################

    #################################################################
    #### Isomorphisms
    #### DEPRECATED!!!! USE EmbeddedLevelGraphs instead!
    ####
    #### This only remains for the BIC generation comparison tests.
    #################################################################
    def _init_invariant(self):
        """
        DEPRECATED!! DO NOT USE!!

        Compute a bunch of invariants (as a quick check for not isommorphic).
        Up to now we have:
        * sorted list of genera (by level)
        * number of edges
        * number of levels
        * names and pole orders of marked points
        * sorted list of pole orders
        """
        self._invariant = (
                len(self.edges),
                self.codim(),
                [(l,self.orderatleg(l)) for l in sorted(self.list_markings())],
                sorted([self.orderatleg(l) for l in self.leglist]))
    def invariant(self):
        """
        DEPRECATED!!! DO NOT USE!!!
        """
        self._init_invariant()
        return self._invariant

    def _genisom_preserves_levels(self,other,dV):
        """
        Check if a "vertex isomorphism" preserves the (relative) level structure
        """
        return all([self.sortedlevels.index(self.levelofvertex(sv))
                    == other.sortedlevels.index(other.levelofvertex(ov))
                        for sv,ov in dV.items()])

    def _legisom_preserves_poleorders(self,other,dL):
        """
        Check if a "leg isomorphism" preserves pole orders.
        """
        return all([self.orderatleg(sl)==other.orderatleg(ol)
                    for sl,ol in dL.items()])

    def _leveliso(self,other,dV):
        """
        Give a dictionary translating the levels of self to other in accordance
        with the dictionary of vertices dV.
        """
        return {l : other.levelofvertex(dV[self.verticesonlevel(l)[0]])
                    for l in self.levels}

    def is_isomorphic(self,other,adddata=False):
        """
        DEPRECATED!!! Instead, use EmbeddedLevelGraph!!!

        Check if self is isomorphic to other.
        Return boolean + list of isomorphisms if adddata=True.

        Note that our isomorphisms are stgraph isomorphisms plus a dictionary
        translating the level structure of self into that of other.

        At the moment we use admcycles.GraphIsom and check which of these
        preserve the LevelGraph structure (using the above invariants, though!)
        Probably it would be much faster to reimplement this for LevelGraphs
        as it seems quite redundant as is...
        """
        isoms = GraphIsom(self.stgraph,other.stgraph)
        # check if the stable graph isomorphisms preserve level structure
        # GraphIsoms consist of a tuple (dV,dL), where dV is a dictionary
        # translating the vertices and dL is one for the legs.
        levelisoms = [[dV,dL,self._leveliso(other,dV)] for [dV,dL] in isoms
                            if self._genisom_preserves_levels(other,dV)
                            and self._legisom_preserves_poleorders(other,dL)]
        if adddata:
            return (levelisoms != [], levelisoms)
        else:
            return (levelisoms != [])

    #################################################################
    #### Twist groups
    #################################################################
    def twist_group(self,quiet=True,with_degree=False):
        """
        This should not be needed! The stack factor should be computed
        using AdditiveGenerator.stack_factor!!!

        Calculate the index of the simple twist group inside the twist group.

        with_degree=True: return a tuple (e,g) where
            * e = index of simple twist in twist
            * g = index of twist group in level rotation group

        """
        # horizontal edges => trivial twist group :-)
        if self.is_horizontal():
            return 1
        N = len(self.edges)
        M = FreeModule(ZZ,N)
        # kernel of projections to Z/prong Z
        K = M.submodule([M.basis()[i]*self.prong(e)
                            for i,e in enumerate(self.edges)])
        if not quiet:
            print("kernel:", K)
        # submodule generated by edges crossing level i
        # i.e. image of R^Gamma (level rotation group) in ZZ^edges
        t_basis = [sum([M.basis()[i] for i,e in enumerate(self.edges)
                        if self.crosseslevel(e,self.internal_level_number(l))])
                                for l in range(1,self.numberoflevels())]
        E = M.submodule(t_basis)
        if not quiet:
            print("t-module:", E)
        # simple twist group: lcm of delta(l) edge prongs*t_i
        deltas = [self.delta(l,True) for l in range(1,self.numberoflevels())]
        if not quiet:
            print("deltas:", deltas)
        ll = [lcm([d.prong(e) for e in d.edges]) for d in deltas]
        if not quiet:
            print("lcms:", ll)
        st_basis = [sum([ll[l-1] * M.basis()[i] for i,e in enumerate(self.edges)
                        if self.crosseslevel(e,self.internal_level_number(l))])
                                for l in range(1,self.numberoflevels())]
        S = M.submodule(st_basis)
        if not quiet:
            print("simple twist group:")
            print(S)
            print("Intersection:")
            print(K.intersection(E))
        # K.intersection(E) = Twist group (in ZZ^edges)
        tw_gr = K.intersection(E)
        if with_degree:
            return (S.index_in(tw_gr),tw_gr.index_in(E))
        else:
            return S.index_in(tw_gr)

    #################################################################
    #### Plotting
    #################################################################
    def plot_obj(self): 
        """
        Return a sage graphics object generated from a sage graph.

        TODO: Some things that are still ugly:
        * No legs!!
        * currently the vertices have labels (index,genus)
        * loops are vertical not horizontal
        """
        # Create the underlying graph with edge label prong order.
        G = Graph([self.genera,
                    [(self.vertex(e[0]),self.vertex(e[1]),self.prong(e))
                                        for e in self.edges]],
                    format='vertices_and_edges', loops=True, multiedges=True)
        # unfortunately, sage insists of displaying the vertex name (index)
        # and not only the label (genus), so we display both...
        G.relabel(list(enumerate(self.genera)))
        h={l:[(v,self.genus(v)) for v in self.verticesonlevel(l)]
                                for l in self.levels}
        # at the moment we just display a list of legs and orders at the bottom
        legtext = "Marked points: "
        for l in sorted(set(self.levels)):
            points_at_level = [leg for leg in self.list_markings()
                                        if self.levelofleg(leg) == l]
            if points_at_level == []:
                continue
            legtext += "\nAt level %s: "%l
            for leg in points_at_level:
                legtext += "\n" + self._pointtype(leg)
                legtext += " of order " + repr(self.orderatleg(leg))
        print(legtext)
        return G.plot(edge_labels=True, vertex_size=1000,
                    heights=h, layout='ranked') + text(legtext,(0,-0.1),
                    vertical_alignment='top',axis_coords=True,axes=False)



#!/usr/bin/env python
# This script extract the doctests from the tex source main.tex and writes
# them in doctest.rst so that they can be parsed by the sage doctester.

outfile = open("doctests.rst", "w")
infile = open("admcycles_manual.tex")

outfile.write("admcycles manual doctest\n")
outfile.write("=========================\n\n")
outfile.write(".. linkall\n\n")
outfile.write("This file was automatically generated from main.tex, do not edit.\n\n")
outfile.write("TESTS:\n")

l = infile.readline()
n = 0

in_lstlistings = False
bad_doctest = False

while l:
    if in_lstlistings:
        if "\\end{lstlisting}" in l and not l.startswith("%"):
            assert in_lstlistings
            in_lstlistings = False

            if doctest:
                outfile.write("\nLines %d-%d::\n\n" %(n_start, n))
                outfile.write(doctest)
        else:
            if l.startswith("sage: "):
                if "?" in l:
                    bad_doctest = True
                else:
                    bad_doctest = False
            if not bad_doctest:
                l = l.strip()
                if l:
                    doctest += "    " + l
                else:
                    doctest += "    <BLANKLINE>"
                doctest += "\n"
    elif "\\begin{lstlisting}" in l and not l.startswith("%"):
        in_lstlistings = True
        doctest = ""
        n_start = n


    l = infile.readline()
    n += 1

outfile.close()
infile.close()



from __future__ import absolute_import
from __future__ import print_function

# pylint does not know sage
from sage.structure.sage_object import SageObject # pylint: disable=import-error
from sage.rings.rational_field import QQ # pylint: disable=import-error
from sage.misc.cachefunc import cached_function # pylint: disable=import-error
from sage.misc.flatten import flatten # pylint: disable=import-error

from copy import deepcopy

import admcycles.diffstrata.levelstratum

#######################################################################
#######################################################################
###### Point Dictionaries
#######################################################################
## * Points on generalised strata are of the form (i,j) where i is the
##      index of the component and j is the index in the signature of
##      the component.
## * An EmbeddedLevelGraph is a LevelGraph together with a dictionary
##      mapping all non-half-edges (i.e. marked points) to the the points
##      of the associated stratum.
##      Note that the numbering of points starts with 1 while the indices
##      of the signatures start with 0. I.e. in the case of a gen. stratum
##      with only one component, the dictionary is i -> (0,i-1).
## * For clutching, we need dictionaries that map points of strata to
##      points of other strata. This allows us to embed clutched LevelGraphs
##      by composition.
## * When splitting a BIC into top and bottom, we consequently need to
##      remember the dictionary that maps marked points of the top and
##      bottom  gen. stratum into the points of stratum the BIC was
##      embedded into.
##
## Thus, a splitting of a BIC inside stratum X must result in two generalised 
## strata, top and bottom, as well as three dictionaries:
##      * emb_top: points of top -> points of X
##      * emb_bot: points of bottom -> points of X
##      * clutching: points of top -> points of bottom
## All of these dictionaries are injective, the keys of emp_top and clutching
## give a partition of the points of top, the keys of emb_bot and the values
## of clutching give a partition of the points of bottom and the values of
## emb_top and emb_bot give a partition of the points of X.
#######################################################################
#######################################################################

def clutch(X, top, bottom, clutch_dict, emb_dict_top, emb_dict_bot,
            middle=None, clutch_dict_lower={}, clutch_dict_long={}, emb_dict_mid={}):
    """
    Clutch EmbeddedLevelGraphs top and bottom to get a new EmbeddedLevelGraph.

    If middle is specified, we do a "double clutching" top - middle - bottom to
    give a graph in X.
    NOTE: 
        * this cannot be split into two steps, as the intermediate graph is
            not a graph in X!
        * We include the possibility to clutch points in the top component *past*
            the middle to the bottom component (long edges!)

    This is the inverse procedure to EmbeddedLevelGraph.split() and 
    GeneralisedStratum.doublesplit().

    Note that if top or bottom is None, we return None.
    
    Args:
        X (GeneralisedStratum): Stratum that the clutched graph will live in
        top (EmbeddedLevelGraph): Top component to be clutched
        bottom (EmbeddedLevelGraph): Bottom component to be clutched
        middle (EmbeddedLevelGraph, optional, defaults to None): Middle component to
            be clutched.
        clutch_dict (dict): The dictionary giving the clutching information,
            (i.e. the edges that will be inserted).
            In the case of a "simple clutch", i.e. top to bottom with no middle,
            a dictionary:
                point on top stratum -> point on bottom stratum
            In the case of a middle graph, i.e. a "double clutching":
                point on top stratum -> point on middle stratum.
            In this case, we also need clutch_dict_lower for the other clutching
            and clutch_dict_long for the long edges.
        clutch_dict_lower (dict, optional): In the case of a "double clutching", the
            info for the "lower" clutching, i.e. a dictionary
                point on middle stratum -> point on bottom stratum
        clutch_dict_long (dict, optional): In the case of a "double clutching", the
            info for the "long edges", i.e. a dictionary
                point on top stratum -> point on bottom stratum
        emb_dict_top (dict): dictionary giving the embedding of the resulting
            EmbeddedLevelGraph, i.e. points on top stratum -> point on
            enveloping stratum.
        emb_dict_bot (dict): dictionary giving the embedding of the resulting
            EmbeddedLevelGraph, i.e. points on bottom stratum -> point on
            enveloping stratum.
        emb_dict_mid (dict, optional): dictionary giving the embedding of the 
            resulting EmbeddedLevelGraph, i.e. points on middle stratum -> point on
            enveloping stratum.
  
    Returns:
        EmbeddedLevelGraph: Clutched LevelGraph together with embedding.
    
    Returns:
        tuple: EmbeddedLevelGraph: as above
    
    EXAMPLES ::

        sage: from admcycles.diffstrata import *
        sage: X=GeneralisedStratum([Signature((1,1))])
        sage: assert clutch(**X.bics[1].split()).is_isomorphic(X.bics[1])
        sage: assert all(clutch(**B.split()).is_isomorphic(B) for B in X.bics)

        The same works for 3-level graphs and doublesplit:

        sage: X=GeneralisedStratum([Signature((1,1))])
        sage: assert all(X.lookup_graph(*ep).is_isomorphic(clutch(**X.doublesplit(ep))) for ep in X.enhanced_profiles_of_length(2))
        sage: X=GeneralisedStratum([Signature((2,2,-2))])
        sage: assert all(X.lookup_graph(*ep).is_isomorphic(clutch(**X.doublesplit(ep))) for ep in X.enhanced_profiles_of_length(2))  # long time (4 seconds)
    """
    if top is None or bottom is None:
        return None
    ## Build new EmbeddedLevelGraph:
    ## This is somewhat similar to levelstratum's unite_embedded_graphs
    ## but here the result is embedded into a _different_ stratum!
    # In case top, middle or bottom are LevelStrata, we replace them by their smooth
    # component (this is mainly to allow clutch to eat the output of split...)
    if isinstance(top, admcycles.diffstrata.levelstratum.LevelStratum):
        top = top.smooth_LG
    if isinstance(bottom, admcycles.diffstrata.levelstratum.LevelStratum):
        bottom = bottom.smooth_LG
    if isinstance(middle, admcycles.diffstrata.levelstratum.LevelStratum):
        # in particular, middle is not None!
        middle = middle.smooth_LG
    # genera simply get combined
    if middle is None:
        newgenera = top.LG.genera + bottom.LG.genera
    else:
        newgenera = top.LG.genera + middle.LG.genera + bottom.LG.genera
    # We renumber the levels consecutively starting at 0.
    top_levels = top.LG.numberoflevels()
    top_level_dict = {top.LG.internal_level_number(l) : -l for l in range(top_levels)}
    newlevels = [top_level_dict[l] for l in top.LG.levels]
    if middle is None:
        mid_levels = 0
        mid_level_dict = {}
    else:
        # the levels of the middle component have to be shifted:
        mid_levels = middle.LG.numberoflevels()
        mid_level_dict = {middle.LG.internal_level_number(l) : -l - top_levels
                                for l in range(mid_levels)}
        newlevels += [mid_level_dict[l] for l in middle.LG.levels]
    bot_levels = bottom.LG.numberoflevels()
    # The levels of the bottom component have to be shifted by both top and middle:
    bot_level_dict = {bottom.LG.internal_level_number(l) : -l - top_levels - mid_levels
                                for l in range(bot_levels)}
    newlevels += [bot_level_dict[l] for l in bottom.LG.levels]
    # TODO: What is a sensible level dictionary?
    # At the moment, we choose the identity.
    newdlevels = {-l : -l for l in range(top_levels + mid_levels + bot_levels)}
    # The legs have to be renumbered in analogy to levelstratum's unite_embedded_graphs:
    newlegs = []
    newedges = []
    newpoleorders = {}
    newdmp = {}  # the new embedding is given by the emb_dicts
    # For adding the new edges, we have to record the new leg numbers and this will
    # happen at different stages (on each component).
    # The idea is, for each set of edges, to create a list of the form
    # [(starting leg 1, starting leg 2,....), (ending leg 1, ending leg 2,...)]
    # and then zip these lists to create the edges.
    # The points are currently stored (as points on the respective stratum) in the 
    # various clutching dictionaries.
    # We want to move them to lists of the form described above.
    if middle is None:
        # If this is a "simple clutch", we only have to add edges between top
        # and bottom.
        clutch_edges = []  # the new edges to be added
    else:
        # Otherwise, we have to distinguish between:
        # * edges from top to middle:
        clutch_edges_tm = []
        # * edges from middle to bottom:
        clutch_edges_mb = []
        # * and edges from top to bottom (going past middle):
        clutch_edges_tb = []
    leg_number_shift = 0  # the shift for renumbering legs
    legs = 0  # leg counter
    # We now go through the components, starting with the top:
    #
    # Note that we also have to choose the appropriate embedding dictionary.
    # We will also have to record the clutching information, as the points are all 
    # going to be renamed, so we have to remember who will be clutched.
    #
    if middle is None:
        # in the "simple clutch" case, it makes sense to also pass
        # the clutch points to each component; these are stored in clutch_dict
        comp_data = ((top,emb_dict_top,clutch_dict.keys()), 
                     (bottom, emb_dict_bot, clutch_dict.values()))
    else:
        # We have to sort out the clutching in the loop anyway, so we just pass a dummy.
        comp_data = ((top, emb_dict_top, {}),
                     (middle, emb_dict_mid, {}),
                     (bottom, emb_dict_bot, {}))
    # emb_g: graph to be clutched
    # emb_dict: dictionary 
    #   marked points of the enveloping stratum of emb_g -> points of X
    for emb_g, emb_dict, clutch_points in comp_data:
        leg_dict = {}  # old number -> new number
        for i, l in enumerate(flatten(emb_g.LG.legs)):
            # the point numbers on bottom component are shifted by the top number
            newlegnumber = leg_number_shift + i + 1
            leg_dict[l] = newlegnumber
            # while we're at it, we add the pole orders:
            newpoleorders[newlegnumber] = emb_g.LG.poleorders[l]
            legs += 1
            # For the dictionary of marked points (for the embedding), we
            # must distinguish if this is a marked point or a half-edge.
            # Marked points are simply the ones for which we have a key
            # in emb_dict :-)
            # The newdmp must then map this point to the image of its dmp in
            # the new stratum, i.e. the image under emb_dict:
            # Note that emb_dict eats stratum points, i.e. the image of the 
            # leg under dmp.
            try:
                newdmp[newlegnumber] = emb_dict[emb_g.dmp[l]]
            except KeyError:
                pass
        leg_number_shift += legs
        # build (nested) list of legs:
        newlegs += [[leg_dict[l] for l in comp] for comp in emb_g.LG.legs]
        # finally, the edges are renumbered accordingly:
        newedges += [(leg_dict[e[0]], leg_dict[e[1]]) for e in emb_g.LG.edges]
        # We now record the clutching information:
        # Note that the points to be clutched are points on the Stratum!
        #
        # Because everything was renamed with leg_dict, we need to *now* remember
        # the edges to clutch (see discussion above).
        #
        # For this, we now have to distinguish if we're working with
        # two or three components: 
        # * In the two component case, the points to be clutched are
        #   the keys (top) and values (bottom) of clutch_dict.
        #   Said differently: clutch_dict encodes the edges that will be added.
        if middle is None:
            # We save "half" of the clutch points in each step, afterwards we just
            # have to combine these lists to obtain all new edges.
            clutch_edges.append(tuple(leg_dict[emb_g.dmp_inv[p]] for p in clutch_points))
        # * In the three component case, 
        #       * the *keys* of clutch_dict are points of the *top* component
        #       * the *values* clutch_dict are those points on the *middle* component that
        #           are clutched to the *top* component
        #       * the *keys* of clutch_dict_lower are points of the *middle* component
        #       * the *values* of clutch_dict_lower are points of the *bottom* component that
        #           are clutched to the *middle* component
        #       * the *keys* of clutch_dict_long are points of the *top* component
        #       * the *values* of clutch_dict_long are points of the *bottom* component that
        #           are clutched to the *top* component (long edges going past the middle)
        #   Said differently:
        #       * clutch_dict encodes edges top - middle
        #       * clutch_dict_lower encodes edges middle - bottom
        #       * clutch_dict_long encodes edges top - bottom (long edges going past middle)
        else:
            # We save the tuple of legs on the current component:
            if emb_g is top:
                clutch_edges_tm.append(tuple(
                    leg_dict[emb_g.dmp_inv[p]] for p in clutch_dict.keys()
                ))
                clutch_edges_tb.append(tuple(
                    leg_dict[emb_g.dmp_inv[p]] for p in clutch_dict_long.keys()
                ))
            elif emb_g is middle:
                clutch_edges_tm.append(tuple(
                    leg_dict[emb_g.dmp_inv[p]] for p in clutch_dict.values()
                ))
                clutch_edges_mb.append(tuple(
                    leg_dict[emb_g.dmp_inv[p]] for p in clutch_dict_lower.keys()
                ))
            else:
                assert emb_g is bottom
                clutch_edges_tb.append(tuple(
                    leg_dict[emb_g.dmp_inv[p]] for p in clutch_dict_long.values()
                ))
                clutch_edges_mb.append(tuple(
                    leg_dict[emb_g.dmp_inv[p]] for p in clutch_dict_lower.values()
                ))
    # Now the actual clutching happens, i.e. adding in the extra edges between  
    # the points of clutch_dict:
    if middle is None:
        assert len(clutch_edges) == 2
        newedges += list(zip(*clutch_edges))
    else:
        assert len(clutch_edges_tm) == 2
        assert len(clutch_edges_tb) == 2
        assert len(clutch_edges_mb) == 2
        newedges += list(zip(*clutch_edges_tm)) + list(zip(*clutch_edges_tb)) + list(zip(*clutch_edges_mb))

    LG = admcycles.diffstrata.levelgraph.LevelGraph(
        newgenera, newlegs, newedges, newpoleorders, newlevels
    )

    ELG = admcycles.diffstrata.levelstratum.EmbeddedLevelGraph(X, LG, newdmp, newdlevels)

    if not ELG.is_legal():
        # This should not happen if R-GRC is checked!
        # print("Illegal clutch attempted: %r, %r" % (top,bottom))
        raise RuntimeError("Illegal clutch, this should not happen anymore!")

    return ELG

class GenDegenerationGraph(SageObject):
    """
    The degeneration graph of the Generalised Stratum X.

    A degeneration graph of a (generalised) stratum is built recursively.
    For that we need to know:
        * The BICs inside X
        * For each BIC its top and bottom component (generalised strata)
        * For each BIC the clutching map of these components
        * For each top and bottom component of each BIC an embedding of their 
            respective BICs into BIC(X)
    This allows us to translate products of BICs on X into clutchings of
    products of BICs on top and bottom components and gives a recursive
    procedure to compute the LevelGraph(s) for any product of BICs in X.
    
    Args:
        X (GeneralisedStratum): The GeneralisedStratum we give the degeneration
            graph of.
    """
    def __init__(self,X):
        self.X = X
        self.n = len(self.X.bics)  # possibly generates all bics of X (!)
        self._top_to_bic = [None for _ in range(self.n)]  # initialise on demand
        self._bot_to_bic = [None for _ in range(self.n)]  # initialise on demand
        self._middle_to_bic = {}  # initialise on demand
        self._top_to_bic_inv = [None for _ in range(self.n)]  # initialise on demand
        self._bot_to_bic_inv = [None for _ in range(self.n)]  # initialise on demand
        self._middle_to_bic_inv = {}  # initialise on demand
        # Note that while top_to_bic and bot_to_bic aren't injective, their
        # images are disjoint, i.e. we can remove the guys we found for top
        # from the ones we're searching through for bot (and vice versa).
        # This happens for every bic, though.
        self._unused_bics_top = [[i for i in range(self.n)] for _ in range(self.n)]
        self._unused_bics_bot = [[i for i in range(self.n)] for _ in range(self.n)]
        # To build all LevelGraphs inside a (generalised) stratum, we store all
        # tuples of indices of bics that give a (non-empty!) 3-level graph:
        # Note that the multiplicities tell us something about automorphisms
        # so we keep track of them and the set in parallel (use _append_triple_prod() !!)
        self._triple_prods = []
        self._triple_prod_set = set()
        self._triple_prod_set_list = None

    def __repr__(self):
        return "GenDegenerationGraph(X=%r)" % self.X
    
    def _append_triple_prod(self,t):
        ## append the tuple t to _triple_prods (both the list and the set...)
        self._triple_prods.append(t)
        self._triple_prod_set.add(t)
    
    @property
    def triple_prods(self):
        """
        A list of profiles of the 3-level graphs in X.

        Note that this is generated on first call.
        
        Returns:
            list: list of tuples of length two of indices of X.bics.
        """
        if self._triple_prod_set_list is None:
            # the easiest way to initialise is to initialise all bic maps:
            for i in range(self.n):
                self.top_to_bic(i)
                self.bot_to_bic(i)
            self._triple_prod_set_list = list(self._triple_prod_set)
        return self._triple_prod_set_list
    
    def top_to_bic(self,i):
        """
        A dictionary indices of BICs of X.bic[i].top -> indices of X.bics.

        Note that this is created once and then reused.
        
        Args:
            i (int): Index of X.bics
        
        Raises:
            RuntimeError: Raised if the squish of the clutching of a degeneration of 
                the top component with X.bic[i].bot is not found in X.bics.
        
        Returns:
            dict: Dictionary mapping BICs of X.bic[i].top to indices of X.bics.
        """
        if self._top_to_bic[i] is None:
            B = self.X.bics[i]
            top_to_bic = {}  # index of B.top.bics -> index of X.bics
            for j, Bt in enumerate(B.top.bics):
                # Note that Bt is embedded (as a Graph) into B.top
                # B.top is embedded (as a stratum) into B (via B.emb_top)
                G = clutch(self.X,Bt,B.bot,B.clutch_dict,B.emb_top,B.emb_bot)
                if G is None:
                    # illegal clutch:
                    raise RuntimeError("Illegal clutch, this should not happen anymore!")
                # Because the degeneration is clutched on top, squishing the
                # bottom level gives a new divisor (i.e. delta(1))
                squished_bic = G.delta(1)
                for im_j in self._unused_bics_top[i]:
                    if im_j == i: 
                        continue  # B^2 is empty
                    if squished_bic.is_isomorphic(self.X.bics[im_j]):
                        top_to_bic[j] = im_j
                        # Record profile of the 3-level graph:
                        self._append_triple_prod((im_j,i))
                        if self._bot_to_bic[i] is None:
                            # otherwise bot_to_bic has already been constructed and
                            # this is futile...
                            try:
                                self._unused_bics_bot[i].remove(im_j)
                            except ValueError:
                                # this is a case where the map is non-injective
                                pass
                        break
                else:  # no break
                    raise RuntimeError("delta(1) of %s, %s, not found in list of BICs of %s"
                        % (G,squished_bic,self.X)
                    )
            self._top_to_bic[i] = top_to_bic.copy()
        return self._top_to_bic[i]

    def bot_to_bic(self,i):
        """
        A dictionary indices of BICs of X.bic[i].bot -> indices of X.bics.

        Note that this is created once and then reused.
        
        Args:
            i (int): Index of X.bics
        
        Raises:
            RuntimeError: Raised if the squish of the clutching of a degeneration of 
                the bottom component with X.bic[i].top is not found in X.bics.
        
        Returns:
            dict: Dictionary mapping BICs of X.bic[i].bot to indices of X.bics.
        """
        if self._bot_to_bic[i] is None:
            B = self.X.bics[i]
            bot_to_bic = {}  # index of B.top.bics -> index of X.bics
            for j, Bb in enumerate(B.bot.bics):
                # Note that Bb is embedded (as a Graph) into B.bot
                # B.bot is embedded (as a stratum) into B (via B.emb_bot)
                G = clutch(self.X,B.top,Bb,B.clutch_dict,B.emb_top,B.emb_bot)
                if G is None:
                    # illegal clutch:
                    raise RuntimeError("Illegal clutch, this should not happen anymore!")
                # Because the degeneration is clutched on bottom, squishing the
                # top level gives a new divisor (i.e. delta(2))
                squished_bic = G.delta(2)
                for im_j in self._unused_bics_bot[i]:
                    if im_j == i: 
                        continue  # B^2 is empty
                    if squished_bic.is_isomorphic(self.X.bics[im_j]):
                        bot_to_bic[j] = im_j
                        # Record profile of the 3-level graph:
                        self._append_triple_prod((i,im_j))
                        if self._top_to_bic[i] is None:
                            # otherwise top_to_bic has already been constructed and
                            # this is futile...
                            try:
                                self._unused_bics_top[i].remove(im_j)
                            except ValueError:
                                # this is a case where the map is non-injective
                                pass
                        break
                else:  # no break
                    raise RuntimeError("delta(2) of %s, %s, not found in list of BICs of %s"
                        % (G,squished_bic,self.X)
                    )
            self._bot_to_bic[i] = bot_to_bic.copy()
        return self._bot_to_bic[i]
    
    def middle_to_bic(self,enh_profile):
        """
        A dictionary for each 3-level graph mapping (indices of) BICs of the middle level
        to (indices of) BICs of the enveloping stratum.
        
        Args:
            enh_profile (tuple): Enhanced profile of a 3-level graph.
        
        Raises:
            RuntimeError: Raised if enh_profile is not a 3-level graph.
            RuntimeError: Raised if a degeneration of the middle level is isomorphic
                    (after squishing) to more than one BIC of X.
            RuntimeError: Raised if no isomorphic BIC of X is found.
        
        Returns:
            dict: dictionary: indices of bics of the middle level -> indices of X.bics
        
        EXAMPLES ::

        """
        if enh_profile not in self._middle_to_bic:
            p, i = enh_profile
            if len(p) != 2:
                raise RuntimeError("Only 3-level graphs have a well-defined middle! %r" % (enh_profile,))
            # The dictionary we want to create:
            mid_to_bic = {}
            # The easiest thing is to split the graph and reuse all the clutching
            # info this yields, replacing the middle level by one of its BICs:
            clutching_info = self.X.doublesplit(enh_profile)
            L = clutching_info['middle']
            # candidates are those BICs that appear as bottom degenerations
            # of the top level *and* as top degenerations of the bottom level:
            candidates = set()
            top_deg = set(self.top_to_bic(p[1]).values())
            for b in self.bot_to_bic(p[0]).values():
                if b in top_deg:
                    candidates.add(b)
            # We now clutch in degenerations of the middle level:
            for i, B in enumerate(L.bics):
                clutching_info['middle'] = B
                H = clutch(**clutching_info)
                # We now apply to delta to find the new BIC.
                # Note that delta is shifted with regard to the profile numbering.
                # It also allows us to recover p:
                assert H.delta(1).is_isomorphic(self.X.bics[p[0]])
                assert H.delta(3).is_isomorphic(self.X.bics[p[1]])
                # delta(2) is the one we are interested in:
                XB = H.delta(2)
                for b in candidates:
                    if XB.is_isomorphic(self.X.bics[b]):
                        # check if more than one is found
                        if i in mid_to_bic:
                            raise RuntimeError("BIC %r of %r seems to be isomorphic to several BICs of %r!" % (i,L,self.X))
                        mid_to_bic[i] = b
                if i not in mid_to_bic:
                    raise RuntimeError("BIC %r of %r not found in BICs of %r!" % (i,L,self.X))
            self._middle_to_bic[enh_profile] = mid_to_bic.copy()
        return self._middle_to_bic[enh_profile]
    
    def top_to_bic_inv(self,i):
        """
        Inverse of top_to_bic.

        More Precisely: A dictionary assigning indices of X.bics a list of indices of
        X.top.bics. Note that this can be more than one (if the intersection is not
        irreducible).

        Note also that not all indices of X.bics are keys (but if not, they should be
        keys of bot_to_bic_inv)!
        
        Args:
            i (int): index of X.bics
        
        Returns:
            dict: index of X.bics -> list of indices of X.top.bics.
        """
        if self._top_to_bic_inv[i] is None:
            self._top_to_bic_inv[i] = {}
            d = self.top_to_bic(i)  # possibly built here (!)
            for j, im_j in d.items():
                try: 
                    self._top_to_bic_inv[i][im_j].append(j)
                except KeyError:
                    self._top_to_bic_inv[i][im_j] = [j]
        return self._top_to_bic_inv[i]

    def bot_to_bic_inv(self,i):
        """
        Inverse of bot_to_bic.

        More Precisely: A dictionary assigning indices of X.bics a list of indices of
        X.bot.bics. Note that this can be more than one (if the intersection is not
        irreducible).

        Note also that not all indices of X.bics are keys (but if not, they should be
        keys of top_to_bic_inv)!
        
        Args:
            i (int): index of X.bics
        
        Returns:
            dict: index of X.bics -> list of indices of X.bot.bics.
        """
        if self._bot_to_bic_inv[i] is None:
            self._bot_to_bic_inv[i] = {}
            d = self.bot_to_bic(i)  # possibly built here (!)
            for j, im_j in d.items():
                try: 
                    self._bot_to_bic_inv[i][im_j].append(j)
                except KeyError:
                    self._bot_to_bic_inv[i][im_j] = [j]
        return self._bot_to_bic_inv[i]

    def middle_to_bic_inv(self,enh_profile):
        """
        Inverse of middle_to_bic.

        More Precisely: A dictionary assigning indices of X.bics a list of indices of
        {middle level of enh_profile}.bics. 
        Note that this can be more than one (if the intersection is not
        irreducible).

        Args:
            enh_profile (tuple): enhanced profile of a 3-level graph.
        
        Returns:
            dict: index of X.bics -> list of indices of {middle level of enh_profile}.bics.
        
        EXAMPLES ::

        """
        if enh_profile not in self._middle_to_bic_inv:
            self._middle_to_bic_inv[enh_profile] = {}
            d = self.middle_to_bic(enh_profile)  # possibly built here (!)
            for j, im_j in d.items():
                try: 
                    self._middle_to_bic_inv[enh_profile][im_j].append(j)
                except KeyError:
                    self._middle_to_bic_inv[enh_profile][im_j] = [j]
        return self._middle_to_bic_inv[enh_profile]



from sage.all import cached_function, PermutationGroup, Partitions, prod
from admcycles import psiclass, lambdaclass, list_strata

# ASV forks before running a benchmark so all caches are empty when a benchmark
# function is first run. However, asv does not fork between repetitions of a
# benchmark so make sure to always run asv with --quick. We could fix this by
# decorating each benchmark or suite such that it forks again; wouldn't be hard
# if we believe that this makes sense.
@cached_function
def nothing_is_cached():
    return None

def to_exp_dict(p):
    d = {}
    for i in p:
        if i in d:
            d[i] += 1
        else:
            d[i] = 1
    return d

class NoProduct:
    r"""
    When a list of .params starts with this entry, not all combinations of the
    entries are processed but each entry is understood as a list of parameters.
    """
    def __init__(self, names):
        self._names = names

    def __repr__(self):
        return self._names

class Benchmark:
    r"""
    Base class for benchmarks that understand the NoProduct marker defined above.
    """
    def setup(self, *args):
        if args and isinstance(args[0], NoProduct):
            raise NotImplementedError("this marker input will be ignored by ASV because we are throwing a NotImplementedError in setup()")

class StrataGeneration(Benchmark):
    def time_list_strata(self, gn):
        g, n = gn
        list_strata(g, n, 3*g-3+n)

    time_list_strata.params = [NoProduct('g, n'), (0, 7), (1, 5), (2, 3), (3, 2), (4, 0)] # (2, 4) times out

class Evaluation(Benchmark):
    def time_psi_classes(self, gn):
        g, n = gn
        for p in Partitions(3*g-3+n, max_length=n):
            prod(psiclass(i+1, g, n)**j for i,j in enumerate(p)).evaluate()

    time_psi_classes.params = [NoProduct('g, n'), (0, 33), (1, 18), (2, 16), (6, 10), (9, 7), (13, 5), (20, 4)]

class AlgebraInTautologicalRing(Benchmark):
    def time_lambda_and_psi_product1(self):
        # from https://gitlab.com/jo314schmitt/admcycles/issues/28
        lambdaclass(3,3,3) * psiclass(1,3,3)**2 * psiclass(2,3,3)**2 * psiclass(3,3,3)**2

    def time_lambda_power(self, dgnp):
        d, g, n, p = dgnp
        lambdaclass(d,g,n)**p

    time_lambda_power.params = [NoProduct('d, g, p, n'), (1, 3, 2, 3), (2, 4, 0, 2)]



from sage.rings.integer_ring import ZZ # pylint: disable=import-error

class Signature(object):
    """
    A signature of a stratum.
    
    Attributes:
        sig (tuple): signature tuple
        g (int): genus
        n (int): total number of points
        p (int): number of poles
        z (int): number of zeroes
        poles (tuple): tuple of pole orders
        zeroes (tuple): tuple of zero orders
        pole_ind (tuple): tuple of indices of poles
        zero_ind (tuple): tuple of indices of zeroes
    
    EXAMPLES ::

        sage: from admcycles.diffstrata.sig import Signature
        sage: sig=Signature((2,1,-1,0))
        sage: sig.g
        2
        sage: sig.n
        4
        sage: sig.poles
        (-1,)
        sage: sig.zeroes
        (2, 1)
        sage: sig.pole_ind
        (2,)
        sage: sig.zero_ind
        (0, 1)
        sage: sig.p
        1
        sage: sig.z
        2
    """
    def __init__(self,sig):
        """
        Initialise signature
        
        Args:
            sig (tuple): signature tuple of integers adding up to 2g-2
        """
        self.sig = sig
        sum_sig = sum(sig)
        if sum_sig % 2 != 0:
            raise ValueError("Error! Illegal signature: Genus not an integer")
        self.g = int(sum_sig/2) + 1
        self.n = len(sig)
        self.poles = tuple(p for p in sig if p < 0)
        self.zeroes = tuple(z for z in sig if z > 0)
        self.marked_points = tuple(k for k in sig if k == 0)
        self.p = len(self.poles)
        self.z = len(self.zeroes)
        self.pole_ind = tuple(i for i,p in enumerate(sig) if p < 0)
        self.zero_ind = tuple(i for i,z in enumerate(sig) if z > 0)

    def __repr__(self):
        return "Signature(%r)" % (self.sig,)
    def __hash__(self):
        return hash(self.sig)
    def __eq__(self,other):
        try:
            return self.sig == other.sig
        except AttributeError:
            return False
